# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import base64
import collections
import copy
import dataclasses
import functools
import itertools
import json
import logging
import math
import re
import zlib
from typing import Any, Callable, Optional, Sequence, Union

import numpy as np
import opt_einsum

import cuequivariance as cue  # noqa: F401
from cuequivariance.etc.linalg import round_to_rational, round_to_sqrt_rational
from cuequivariance.etc.permutations import (
    generate_permutations_from,
    inverse_permutation,
)
from cuequivariance.segmented_polynomials import Path, Subscripts

from .dimensions_dict import format_dimensions_dict

logger = logging.getLogger(__name__)


@dataclasses.dataclass(init=False, frozen=True)
class SegmentedTensorProduct:
    """
    Irreps-agnostic and dataflow-agnostic descriptor of a segmented tensor product

    Args:
        operands (list of operands): The operands of the tensor product. To each operand corresponds subscripts and a list of segments.
        paths (list of paths): Each path contains coefficients and a list of indices.
            The indices are the indices of the segments of the operands.
        coefficient_subscripts (str): The subscripts of the coefficients.


    We typically use the :func:`from_subscripts <cuequivariance.SegmentedTensorProduct.from_subscripts>` method to create a descriptor and then add segments and paths one by one.

    .. rubric:: Methods
    """

    operands_and_subscripts: tuple[tuple[cue.SegmentedOperand, str], ...]
    coefficient_subscripts: str
    paths: tuple[Path, ...]

    ################################ Initializers ################################

    # From here we can use object.__setattr__ to modify the attributes
    def __init__(
        self,
        operands_and_subscripts: Sequence[tuple[cue.SegmentedOperand | None, str]]
        | None = None,
        coefficient_subscripts: str = "",
        *,
        paths: Sequence[Path] | None = None,
    ):
        if operands_and_subscripts is None:
            operands_and_subscripts = []
        if paths is None:
            paths = []

        operands_and_subscripts = tuple(
            (
                ope.copy() if ope is not None else cue.SegmentedOperand(ndim=len(ss)),
                str(ss),
            )
            for ope, ss in operands_and_subscripts
        )

        object.__setattr__(self, "operands_and_subscripts", operands_and_subscripts)
        object.__setattr__(self, "coefficient_subscripts", coefficient_subscripts)
        object.__setattr__(self, "paths", tuple(paths))

    def set_operand(self, oid: int, operand: cue.SegmentedOperand):
        assert oid < len(self.operands_and_subscripts)
        object.__setattr__(
            self,
            "operands_and_subscripts",
            self.operands_and_subscripts[:oid]
            + ((operand.copy(), self.operands_and_subscripts[oid][1]),)
            + self.operands_and_subscripts[oid + 1 :],
        )

    def set_paths(self, paths: list[Path]):
        # no need to deepcopy, because Path is immutable
        object.__setattr__(self, "paths", tuple(paths))

    def insert_path_(self, path_index: int, path: Path):
        # no need to deepcopy, because Path is immutable
        object.__setattr__(
            self, "paths", self.paths[:path_index] + (path,) + self.paths[path_index:]
        )

    # until here. Below we use dataclasses.replace or the setters to modify the attributes

    @property
    def operands(self) -> tuple[cue.SegmentedOperand, ...]:
        return tuple(ope for ope, _ in self.operands_and_subscripts)

    def assert_valid(self):
        assert Subscripts.is_valid(self.subscripts)

        for m in self.subscripts.modes():
            if self.subscripts.count(m) == 1:
                raise ValueError(
                    f"mode {m} is not contracted in {self.subscripts}. It should appear at least twice."
                )

        for operand in self.operands:
            operand.assert_valid()  # check subscripts and segment lengths

        for path in self.paths:
            path.assert_valid()

            if path.coefficients.ndim != len(self.coefficient_subscripts):
                raise ValueError(  # pragma: no cover
                    f"expected {len(self.coefficient_subscripts)} dimensions for the coefficients, got {path.coefficients.ndim}."
                )

            for oid in range(self.num_operands):
                if path.indices[oid] >= self.operands[oid].num_segments:
                    raise ValueError(  # pragma: no cover
                        f"path has an index {path.indices[oid]} out of bounds for operand {oid}."
                        f" Operand {oid} has only {self.operands[oid].num_segments} segments."
                    )

            dims = self.get_path_dimensions_dict(path, returns_sets=True)
            if any(len(dd) != 1 for dd in dims.values()):
                raise ValueError(  # pragma: no cover
                    f"path has ambiguous dimensions {dims}."
                )

    @classmethod
    def from_subscripts(cls, subscripts: Subscripts) -> SegmentedTensorProduct:
        r"""
        Create a descriptor from a subscripts string.

        Examples:
            >>> d = cue.SegmentedTensorProduct.from_subscripts("uv,ui,vj+ij")
            >>> i0 = d.add_segment(0, (2, 3))
            >>> i1 = d.add_segment(1, (2, 5))
            >>> i2 = d.add_segment(2, (3, 4))
            >>> d.add_path(i0, i1, i2, c=np.ones((5, 4)))
            0
            >>> print(d)
            uv,ui,vj+ij operands=[(2, 3)],[(2, 5)],[(3, 4)] paths=[op0[0]*op1[0]*op2[0]*c c.shape=(5, 4) c.nnz=20]
        """
        subscripts = Subscripts(subscripts)
        operands = [
            cue.SegmentedOperand(ndim=len(operand)) for operand in subscripts.operands
        ]

        return cls(
            operands_and_subscripts=list(zip(operands, subscripts.operands)),
            paths=[],
            coefficient_subscripts=subscripts.coefficients,
        )

    @classmethod
    def empty_segments(cls, num_segments: list[int]) -> SegmentedTensorProduct:
        r"""
        Create a descriptor with a simple structure.

        Examples:
            >>> cue.SegmentedTensorProduct.empty_segments([2, 3, 4])
            ,, sizes=2,3,4 num_segments=2,3,4 num_paths=0
        """
        return cls(
            operands_and_subscripts=[
                (cue.SegmentedOperand.empty_segments(num), "") for num in num_segments
            ],
            paths=[],
            coefficient_subscripts="",
        )

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> SegmentedTensorProduct:
        r"""Create a descriptor from a dictionary."""
        d = SegmentedTensorProduct.from_subscripts(data["subscripts"])
        for oid, operand in enumerate(data["operands"]):
            d.add_segments(oid, operand["segments"])
        for indices, coefficients in zip(
            data["paths"]["indices"], data["paths"]["coefficients"]
        ):
            d.add_path(*indices, c=np.array(coefficients))
        return d

    @classmethod
    def from_json(cls, data: str) -> SegmentedTensorProduct:
        r"""Create a descriptor from a JSON string."""
        return cls.from_dict(json.loads(data))

    @classmethod
    def from_bytes(cls, data: bytes) -> SegmentedTensorProduct:
        r"""Create a descriptor from compressed binary data."""
        return cls.from_json(zlib.decompress(data).decode("ascii"))

    @classmethod
    def from_base64(cls, data: str) -> SegmentedTensorProduct:
        r"""Create a descriptor from a base64 string."""
        return cls.from_bytes(base64.b64decode(data))

    ################################ Properties ################################

    def __hash__(self) -> int:
        return hash(
            (self.operands_and_subscripts, self.paths, self.coefficient_subscripts)
        )

    def __eq__(self, value: SegmentedTensorProduct) -> bool:
        assert isinstance(value, SegmentedTensorProduct)
        return (
            self.operands_and_subscripts == value.operands_and_subscripts
            and self.paths == value.paths
            and self.coefficient_subscripts == value.coefficient_subscripts
        )

    def __lt__(self, value: SegmentedTensorProduct) -> bool:
        assert isinstance(value, SegmentedTensorProduct)
        return (
            self.num_operands,
            self.num_paths,
            self.subscripts,
            self.operands,
            self.paths,
            self.coefficient_subscripts,
        ) < (
            value.num_operands,
            value.num_paths,
            value.subscripts,
            value.operands,
            value.paths,
            value.coefficient_subscripts,
        )

    def __repr__(self) -> str:
        if max(len(operand) for operand in self.operands) == 1 and len(self.paths) == 1:
            operands = ",".join(
                "[" + ",".join(map(str, operand.segments)) + "]"
                for operand in self.operands
            )
            return f"{self.subscripts} operands={operands} paths={list(self.paths)}"
        sizes = ",".join(f"{operand.size}" for operand in self.operands)
        num_segments = ",".join(f"{len(operand)}" for operand in self.operands)
        output = f"{self.subscripts} sizes={sizes} num_segments={num_segments} num_paths={self.num_paths}"
        dims = format_dimensions_dict(self.get_dimensions_dict())
        if dims:
            output += f" {dims}"
        return output

    @property
    def num_operands(self) -> int:
        """Number of operands."""
        return len(self.operands_and_subscripts)

    @property
    def num_paths(self) -> int:
        """Number of paths."""
        return len(self.paths)

    @property
    def subscripts(self) -> Subscripts:
        """Subscripts of the tensor product."""
        return Subscripts.from_operands(
            [subscripts for _, subscripts in self.operands_and_subscripts],
            self.coefficient_subscripts,
        )

    @property
    def indices(self) -> np.ndarray:
        """Indices of the paths."""
        if len(self.paths) == 0:
            return np.empty((0, self.num_operands), dtype=int)
        return np.array([path.indices for path in self.paths], dtype=int)

    @property
    def coefficients_are_stackable(self) -> bool:
        """Check if the coefficients are stackable."""
        if len({path.coefficients.shape for path in self.paths}) == 1:
            return True
        dims = self.get_dimensions_dict()
        if all(
            len(dd) == 1 for ch, dd in dims.items() if ch in self.coefficient_subscripts
        ):
            return True
        return False

    @property
    def stacked_coefficients(self) -> np.ndarray:
        """Stacked coefficients of the paths in a single array."""
        if not self.coefficients_are_stackable:
            raise ValueError("coefficients are not stackable.")  # pragma: no cover

        cs = [path.coefficients for path in self.paths]
        if len(cs) == 0:
            dims = self.get_dimensions_dict()
            shape = tuple(
                next(iter(dims.get(ch))) for ch in self.coefficient_subscripts
            )
            return np.empty((0,) + shape, dtype=np.float64)
        return np.stack(cs)

    ################################ Getters ################################

    def to_text(self, coefficient_formatter=lambda x: f"{x}") -> str:
        """Human-readable text representation of the descriptor.

        Args:
            coefficient_formatter (callable, optional): A function to format the coefficients.

        Examples:
            >>> ((_, d),) = cue.descriptors.fully_connected_tensor_product(
            ...     cue.Irreps("SO3", "4x0+4x1"),
            ...     cue.Irreps("SO3", "4x0+4x1"),
            ...     cue.Irreps("SO3", "4x0+4x1")
            ... ).polynomial.operations
            >>> d = d.flatten_coefficient_modes()
            >>> print(d.to_text())
            uvw,u,v,w sizes=320,16,16,16 num_segments=5,4,4,4 num_paths=16 u=4 v=4 w=4
            operand #0 subscripts=uvw
              | u: [4] * 5
              | v: [4] * 5
              | w: [4] * 5
            operand #1 subscripts=u
              | u: [4] * 4
            operand #2 subscripts=v
              | v: [4] * 4
            operand #3 subscripts=w
              | w: [4] * 4
            Flop cost: 0->1344 1->2368 2->2368 3->2368
            Memory cost: 368
            Path indices: 0 0 0 0, 1 0 1 1, 1 0 2 2, 1 0 3 3, 2 1 0 1, 2 2 0 2, ...
            Path coefficients: [0.17...]
        """
        out = f"{self}"
        dims = self.get_dimensions_dict()
        for oid, (operand, subscripts) in enumerate(self.operands_and_subscripts):
            out += f"\noperand #{oid} subscripts={subscripts}"
            for i, ch in enumerate(subscripts):
                if len(dims[ch]) == 1:
                    out += f"\n  | {ch}: [{operand.segments[0][i]}] * {len(operand.segments)}"
                else:
                    out += (
                        f"\n  | {ch}: ["
                        + ", ".join(str(s[i]) for s in operand.segments)
                        + "]"
                    )

        out += f"\nFlop cost: {' '.join(f'{oid}->{self.flop(oid)}' for oid in range(self.num_operands))}"
        out += f"\nMemory cost: {self.memory([1] * self.num_operands)}"

        if len(self.paths) > 0:
            out += "\nPath indices: " + ", ".join(
                " ".join(str(i) for i in row) for row in self.indices
            )

            if coefficient_formatter is not None:
                formatter = {"float": coefficient_formatter}
                if all(len(dims[ch]) == 1 for ch in self.coefficient_subscripts):
                    out += "\nPath coefficients: " + np.array2string(
                        self.stacked_coefficients,
                        max_line_width=np.inf,
                        formatter=formatter,
                        threshold=np.inf,
                    )
                else:
                    out += "\nPath coefficients:\n" + "\n".join(
                        np.array2string(
                            path.coefficients, formatter=formatter, threshold=np.inf
                        )
                        for path in self.paths
                    )
        else:
            out += "\nNo paths."

        return out

    def to_dict(self, extended: bool = False) -> dict[str, Any]:
        """Dictionary representation of the descriptor."""
        paths = {
            "indices": [path.indices for path in self.paths],
            "coefficients": [path.coefficients.tolist() for path in self.paths],
        }
        if not extended:
            return {
                "subscripts": str(self.subscripts),
                "operands": [{"segments": ope.segments} for ope in self.operands],
                "paths": paths,
            }

        segment_slices = [ope.segment_slices() for ope in self.operands]
        extended_dict = {
            "subscripts": str(self.subscripts),
            "coefficient_subscripts": self.coefficient_subscripts,
            "operands": [
                {
                    "subscripts": ss,
                    "segments": ope.segments,
                    "size": ope.size,
                    "segment_offsets": [sl.start for sl in slices],
                    "segment_sizes": [sl.stop - sl.start for sl in slices],
                    "flops": self.flop(oid),
                }
                for oid, (ope, ss), slices in zip(
                    range(self.num_operands),
                    self.operands_and_subscripts,
                    segment_slices,
                )
            ],
            "memory": self.memory([1] * self.num_operands),
            "paths": paths,
        }
        return extended_dict

    def to_json(self, extended: bool = False) -> str:
        """JSON representation of the descriptor."""
        return json.dumps(self.to_dict(extended))

    def to_bytes(self, extended: bool = False) -> bytes:
        """Compressed binary representation of the descriptor."""
        return zlib.compress(self.to_json(extended).encode("ascii"))

    def to_base64(self, extended: bool = False) -> str:
        """
        Base64 representation of the descriptor.

        Examples:
            >>> d = cue.descriptors.fully_connected_tensor_product(
            ...     cue.Irreps("SO3", "4x0+4x1"),
            ...     cue.Irreps("SO3", "4x0+4x1"),
            ...     cue.Irreps("SO3", "4x0+4x1")
            ... ).polynomial.operations[0][1]
            >>> print(d.to_base64())
            eJytkstuwjAQRX/F8r...lTF2zlX91/fHyvj2Z4=
        """
        return base64.b64encode(self.to_bytes(extended)).decode("ascii")

    @functools.cache
    def get_dimensions_dict(self) -> dict[str, set[int]]:
        """Get the dimensions of the tensor product."""
        dims: dict[str, set[int]] = {ch: set() for ch in self.subscripts.modes()}
        for operand, subscripts in self.operands_and_subscripts:
            for i, dd in operand.get_dimensions_dict().items():
                dims[subscripts[i]].update(dd)
        # Note: no need to go through the coefficients since must be contracted with the operands
        return dims

    def get_dims(self, m: str) -> set[int]:
        """
        Get the dimensions of a specific mode.

        Examples:
            >>> d = cue.descriptors.fully_connected_tensor_product(
            ...     cue.Irreps("SO3", "4x0+8x1"),
            ...     cue.Irreps("SO3", "3x0+3x1"),
            ...     cue.Irreps("SO3", "5x0+7x1")
            ... ).polynomial.operations[0][1]
            >>> d.get_dims("u")
            {8, 4}
            >>> d.get_dims("v")
            {3}
            >>> d.get_dims("w")
            {5, 7}
        """
        return self.get_dimensions_dict().get(m, set())

    def get_path_dimensions_dict(
        self, path: Union[int, Path], *, returns_sets: bool = False
    ) -> dict[str, Union[int, set[int]]]:
        """Get the dimensions of a specific path."""
        if isinstance(path, int):
            path = self.paths[path]

        dims = {
            m: {d} for m, d in zip(self.coefficient_subscripts, path.coefficients.shape)
        }
        for oid, sid in enumerate(path.indices):
            for m, d in zip(self.subscripts.operands[oid], self.operands[oid][sid]):
                dims.setdefault(m, set()).add(d)

        if returns_sets:
            return dims
        assert all(len(dd) == 1 for dd in dims.values())
        return {m: next(iter(dd)) for m, dd in dims.items()}

    def get_path_dim(
        self, path: Union[int, Path], m: str, *, returns_set=False
    ) -> Union[int, set[int]]:
        """Get the dimension of a specific mode in a specific path."""
        if isinstance(path, int):
            path = self.paths[path]
        return self.get_path_dimensions_dict(path, returns_sets=returns_set).get(
            m, set() if returns_set else 0
        )

    def segment_slice(self, operand: int, path: Union[int, Path]) -> slice:
        """Get the slice of the segment in the given operand selected by the given path."""
        if isinstance(path, int):
            path = self.paths[path]
        return self.operands[operand].segment_slices()[path.indices[operand]]

    def get_segment_shape(
        self, operand: int, path: Union[int, Path]
    ) -> tuple[int, ...]:
        """Get the shape of the segment in the given operand selected by the given path."""
        if isinstance(path, int):
            path = self.paths[path]
        return self.operands[operand][path.indices[operand]]

    def all_segments_are_used(self) -> bool:
        """Check if all segments are used in the tensor product."""
        for op_idx, operand in enumerate(self.operands):
            for seg_idx in range(operand.num_segments):
                if not any(seg_idx == path.indices[op_idx] for path in self.paths):
                    return False
        return True

    def compressed_path_segment(self, operand: int) -> np.ndarray:
        """
        Starting indices of paths for the segments of the specified operand.

        Note: This method requires that the paths are sorted by the specified operand.

        Args:
            operand (int): The index of the operand for which to find path groups.

        Returns:
            np.ndarray: An array of starting path indices for each segment in the specified operand.

        Examples:

            .. code-block:: python

                indices[:, operand], operands[operand].num_segments -> compressed_path_segment(operand)
                [0, 0, 1, 1, 1], 2 -> [0, 2, 5]
                [0, 0, 1, 1, 1], 3 -> [0, 2, 5, 5]
                [0, 0, 2, 2, 2], 3 -> [0, 2, 2, 5]
                [1, 1], 2 -> [0, 0, 2]
        """
        if not np.all(np.diff(self.indices[:, operand]) >= 0):
            raise ValueError("Paths must be sorted by the specified operand.")

        i = self.indices[:, operand]
        n = self.operands[operand].num_segments

        return np.append(0, np.bincount(i, minlength=n).cumsum())

    def operands_with_identical_segments(self) -> frozenset[frozenset[int]]:
        """Groups of operands sharing the same segments."""
        operand_to_oid = collections.defaultdict(list)
        for oid, ope in enumerate(self.operands):
            operand_to_oid[ope].append(oid)
        return frozenset(map(frozenset, operand_to_oid.values()))

    @functools.cache
    def symmetries(self) -> list[tuple[int, ...]]:
        """List of permutations that leave the tensor product invariant."""

        def clean(d):
            return d.consolidate_paths().canonicalize_subscripts()

        d = clean(self)

        ps = set()
        for group in self.operands_with_identical_segments():
            group = sorted(group)
            for p in itertools.permutations(range(len(group))):
                p = tuple(
                    group[p[group.index(i)]] if i in group else i
                    for i in range(self.num_operands)
                )
                if p in ps:
                    continue
                if d == clean(d.permute_operands(p)):
                    ps.add(p)
                    ps.add(inverse_permutation(p))
                    ps = generate_permutations_from(ps)
        return sorted(ps)

    def coefficients_equal_one(self) -> bool:
        """Check if all coefficients are equal to one."""
        return np.all(self.stacked_coefficients == 1)

    def flop(
        self, operand: int, batch_size: int = 1, algorithm: str = "optimal"
    ) -> int:
        """
        Compute the number of flops needed to compute the specified operand.

        Args:
            operand (int): The operand for which to compute the flop cost.
            batch_size (int, optional): The batch size for the computation. Defaults to 1.
            algorithm (str, optional): The algorithm to use to compute the cost. Can be 'optimal' or 'naive'.

        Returns:
            int: The number of flops needed to compute the specified operand.
        """
        d = self.move_operand_last(operand)
        subscripts = (
            d.coefficient_subscripts
            + "".join("," + ss for ss in d.subscripts.operands[:-1])
            + "->"
            + d.subscripts.operands[-1]
        )

        @functools.cache
        def compute_cost(segment_shapes: tuple[tuple[int, ...], ...]) -> int:
            _, info = opt_einsum.contract_path(
                subscripts, *segment_shapes, optimize="optimal", shapes=True
            )
            if algorithm == "naive":
                return int(info.naive_cost)
            elif algorithm == "optimal":
                return int(info.opt_cost)
            else:
                raise ValueError(f"unknown algorithm {algorithm}.")

        cost = 0
        for path in d.paths:
            shapes = tuple(
                d.get_segment_shape(oid, path) for oid in range(d.num_operands - 1)
            )
            dims = d.get_path_dimensions_dict(path)
            coeff_shape = tuple(dims[ch] for ch in d.coefficient_subscripts)
            cost += compute_cost((coeff_shape,) + shapes)
        return cost * batch_size

    def memory(self, batch_sizes: list[int]) -> int:
        """Compute the memory usage of the tensor product."""
        assert len(batch_sizes) == self.num_operands
        return sum(Z * ope.size for Z, ope in zip(batch_sizes, self.operands))

    ################################ Modifiers ################################

    def insert_path(
        self,
        path_index,
        *segments: Union[int, tuple[int, ...], dict[str, int], None],
        c: np.ndarray,
        dims: Optional[dict[str, int]] = None,
    ) -> int:
        """Insert a path at a specific index."""
        path_index = _canonicalize_index("path_index", path_index, len(self.paths) + 1)

        if len(segments) != self.num_operands:
            raise ValueError(
                f"expected {self.num_operands} indices, got {len(segments)}."
            )

        coefficients = np.asarray(c)
        del c

        if coefficients.ndim != len(self.coefficient_subscripts):
            raise ValueError(
                f"expected {len(self.coefficient_subscripts)} dimensions for the coefficients, got {coefficients.ndim}."
            )

        if dims is not None:
            dims: dict[str, set[int]] = {
                m: {d} for m, d in dims.items() if m in self.subscripts.modes()
            }
        else:
            dims: dict[str, set[int]] = dict()

        for m, d in zip(self.coefficient_subscripts, coefficients.shape):
            dims.setdefault(m, set()).add(d)

        for oid, s in enumerate(segments):
            subscripts = self.subscripts.operands[oid]
            if isinstance(s, int):
                if not (
                    -self.operands[oid].num_segments
                    <= s
                    < self.operands[oid].num_segments
                ):
                    raise ValueError(
                        f"segment index {s} out of bounds for operand {oid}."
                    )
                for m, d in zip(subscripts, self.operands[oid][s]):
                    dims.setdefault(m, set()).add(d)
            elif isinstance(s, tuple):
                for m, d in zip(subscripts, s):
                    dims.setdefault(m, set()).add(d)
            elif isinstance(s, dict):
                for m, d in s.items():
                    dims.setdefault(m, set()).add(d)
            elif s is None:
                pass
            else:
                raise ValueError(
                    f"expected a segment index, a tuple or a dictionary. Got {s}."
                )

        if any(len(dd) != 1 for dd in dims.values()):
            raise ValueError(
                f"Ambiguous dimensions for the path. {dims} try to insert {segments}"
            )
        if len(dims) != len(self.subscripts.modes()):
            raise ValueError(
                f"expected a dimension for each subscripts in the descriptor. Got {dims}."
                f" Missing dimensions for {self.subscripts.modes() - dims.keys()}."
            )

        dims = {m: next(iter(dd)) for m, dd in dims.items()}

        path = Path(
            [
                (
                    (s + self.operands[oid].num_segments)
                    % self.operands[oid].num_segments
                    if isinstance(s, int)
                    else self.add_segment(oid, dims)
                )
                for oid, s in enumerate(segments)
            ],
            coefficients,
        )
        self.insert_path_(path_index, path)
        return path_index

    def add_path(
        self,
        *segments: Union[int, tuple[int, ...], dict[str, int], None],
        c: np.ndarray,
        dims: Optional[dict[str, int]] = None,
    ) -> int:
        """
        Add a path to the descriptor.

        Args:
            segments: Specifies the segments of the operands that are contracted in the path.
            c (np.ndarray): The coefficients of the path.
            dims (dict[str, int], optional): The extent of the modes.

        Returns:
            int: The index of the added path.

        Examples:
            >>> d = cue.SegmentedTensorProduct.from_subscripts("uv,ui,vj+ij")
            >>> i1 = d.add_segment(1, (2, 3))
            >>> i2 = d.add_segment(2, (2, 5))

            We can use ``None`` to add a new segment on the fly:
            >>> d.add_path(None, i1, i2, c=np.ones((3, 5)))
            0

            The descriptor has now a new segment ``(2, 2)`` in the first operand:
            >>> d
            uv,ui,vj+ij operands=[(2, 2)],[(2, 3)],[(2, 5)] paths=[op0[0]*op1[0]*op2[0]*c c.shape=(3, 5) c.nnz=15]
            >>> d.add_path(0, None, None, c=np.ones((10, 10)))
            1
            >>> d
            uv,ui,vj+ij sizes=4,26,30 num_segments=1,2,2 num_paths=2 i={3, 10} j={5, 10} u=2 v=2

            When the dimensions of the modes cannot be inferred, we can provide them:
            >>> d.add_path(None, None, None, c=np.ones((2, 2)), dims={"u": 2, "v": 2})
            2
            >>> d
            uv,ui,vj+ij sizes=8,30,34 num_segments=2,3,3 num_paths=3 i={2, 3, 10} j={2, 5, 10} u=2 v=2
        """
        return self.insert_path(len(self.paths), *segments, c=c, dims=dims)

    def insert_segments(self, operand: int, sid: int, segments: list[tuple[int, ...]]):
        """Insert segments at a specific index."""
        operand = _canonicalize_index("operand", operand, self.num_operands)
        sid = _canonicalize_index("sid", sid, self.operands[operand].num_segments + 1)

        o = self.operands[operand]
        n = cue.SegmentedOperand(ndim=o.ndim, segments=segments)
        self.set_operand(
            operand,
            cue.SegmentedOperand(
                ndim=o.ndim,
                segments=o.segments[:sid] + n.segments + o.segments[sid:],
                _dims={m: o.get_dims(m) | n.get_dims(m) for m in range(o.ndim)},
            ),
        )
        self.set_paths(
            [
                Path(
                    [
                        s if s < sid or oid != operand else s + n.num_segments
                        for oid, s in enumerate(path.indices)
                    ],
                    path.coefficients,
                )
                for path in self.paths
            ],
        )

    def add_segment(
        self, operand: int, segment: Union[tuple[int, ...], dict[str, int]]
    ) -> int:
        """Add a segment to the descriptor."""
        if isinstance(segment, dict):
            segment = tuple(segment[m] for m in self.subscripts.operands[operand])
        return self.operands[operand].add_segment(segment)

    def add_segments(
        self, operand: int, segments: list[Union[tuple[int, ...], dict[str, int]]]
    ):
        """Add segments to the descriptor."""
        for segment in segments:
            self.add_segment(operand, segment)

    def canonicalize_subscripts(self) -> SegmentedTensorProduct:
        """
        Return a new descriptor with a canonical representation of the subscripts.

        Examples:
            >>> d = cue.SegmentedTensorProduct.from_subscripts("ab,ax,by+yx")
            >>> d.canonicalize_subscripts()
            uv,ui,vj+ij sizes=0,0,0 num_segments=0,0,0 num_paths=0 i= j= u= v=

        This is useful to identify equivalent descriptors.
        """
        subscripts = Subscripts.canonicalize(self.subscripts)
        d = self.add_or_rename_modes(subscripts)
        d = d.add_or_transpose_modes(
            Subscripts.from_operands(
                d.subscripts.operands, "".join(sorted(d.coefficient_subscripts))
            )
        )
        return d

    def add_or_rename_modes(
        self, subscripts: str, *, mapping: Optional[dict[str, str]] = None
    ) -> SegmentedTensorProduct:
        r"""
        Return a new descriptor with the modes renamed according to the new subscripts.

        Args:
            subscripts (str): The new subscripts that contains the new names of the modes.
                The new subscripts can also be a superset of the old subscripts.
            mapping (dict of str to str, optional): The mapping between the old and new modes.

        Returns:
            SegmentedTensorProduct: The new descriptor with the renamed modes.
        """
        subscripts = Subscripts(subscripts)

        if subscripts.is_equivalent(self.subscripts):
            return SegmentedTensorProduct(
                [(ope, ss) for ope, ss in zip(self.operands, subscripts.operands)],
                subscripts.coefficients,
                paths=self.paths,
            )

        # Non trivial setting: the new subscripts might be a superset of the old subscripts
        # In this case, we need to properly map de mode dimensions and put 1s where needed
        # But we need to be careful, because the new subscripts might rename the modes differently
        if mapping is None:
            mappings = self.subscripts.is_subset_of(subscripts)
            if len(mappings) == 0:
                raise ValueError(
                    f"expected new subscripts {subscripts} to contain all the modes of old subscripts {self.subscripts}."
                )
            if len(mappings) > 1:
                raise ValueError(
                    f"unable to determine non-ambiguous mapping between old subscripts {self.subscripts} and new subscripts {subscripts}."
                )
            mapping = mappings[0]

        if not all(ch in mapping for ss in self.subscripts.operands for ch in ss):
            raise ValueError(
                f"expected all segment modes to be in the mapping {mapping}."
            )
        if not all(ch in subscripts for ch in mapping.values()):
            raise ValueError(
                f"expected all mapped modes to be in the new subscripts {subscripts}."
            )

        D = SegmentedTensorProduct.from_subscripts(subscripts)
        for oid in range(self.num_operands):
            for sid in range(self.operands[oid].num_segments):
                dims = collections.defaultdict(lambda: 1)
                for m, d in zip(self.subscripts.operands[oid], self.operands[oid][sid]):
                    dims[mapping[m]] = d
                D.add_segment(oid, dims)

        for path in self.paths:
            dims: dict[str, int] = dict()
            for oid, sid in enumerate(path.indices):
                for m, d in zip(D.subscripts.operands[oid], D.operands[oid][sid]):
                    dims[m] = d

            D.insert_path_(
                len(D.paths),
                Path(
                    indices=path.indices,
                    coefficients=np.reshape(
                        path.coefficients,
                        tuple(dims[m] for m in D.coefficient_subscripts),
                    ),
                ),
            )

        return D

    def add_or_transpose_modes(
        self, subscripts: str, dims: Optional[dict[str, int]] = None
    ) -> SegmentedTensorProduct:
        r"""
        Return a new descriptor with the modes transposed according to the new subscripts.

        Args:
            subscripts (str): A new subscripts that contains a permutation of the modes of the current subscripts.
            dims (dict of str to int, optional): The dimensions of the new modes.

        Returns:
            SegmentedTensorProduct: The new descriptor with the transposed modes.
        """
        subscripts = Subscripts.complete_wildcards(subscripts, self.subscripts)

        if dims is None:
            dims = dict()

        for old, new in zip(
            self.subscripts.operands_and_coefficients,
            subscripts.operands_and_coefficients,
        ):
            if not set(old).issubset(set(new)):
                raise ValueError(f"expected {old} to be a subset of {new}.")

            for ch in new:
                if ch not in old and ch not in dims:
                    raise ValueError(f"expected dimension for mode {ch}.")

        if len(self.subscripts.coefficients) != len(subscripts.coefficients):
            raise ValueError(
                "impossible to introduce new modes in the coefficients subscripts."
            )

        d = SegmentedTensorProduct.from_subscripts(subscripts)
        for oid in range(self.num_operands):
            old, new = (
                self.subscripts.operands[oid],
                d.subscripts.operands[oid],
            )
            perm = [old.index(ch) if ch in old else ch for ch in new]
            for sid in range(self.operands[oid].num_segments):
                d.add_segment(
                    oid,
                    tuple(
                        self.operands[oid][sid][i] if isinstance(i, int) else dims[i]
                        for i in perm
                    ),
                )
        old, new = self.coefficient_subscripts, d.coefficient_subscripts
        perm = [old.index(ch) for ch in new]
        d.set_paths(
            [
                Path(
                    indices=path.indices,
                    coefficients=np.transpose(path.coefficients, perm),
                )
                for path in self.paths
            ]
        )
        return d

    def append_modes_to_all_operands(
        self, modes: str, dims: dict[str, int]
    ) -> SegmentedTensorProduct:
        r"""
        Return a new descriptor with the modes appended (to the right) to all operands.

        Args:
            modes (str): The new segment modes to append to all operands.
            dims (dict of str to int): The dimensions of the new modes.

        Returns:
            SegmentedTensorProduct: The new descriptor with the appended modes.
        """
        if not all(ch not in self.subscripts for ch in modes):
            raise ValueError(
                f"expected new modes {modes} to be disjoint from the current subscripts {self.subscripts}."
            )
        if not all(ch in dims for ch in modes):
            raise ValueError(f"expected dimensions for all new modes {modes}.")
        subscripts = Subscripts.from_operands(
            [ope + modes for ope in self.subscripts.operands],
            coefficients=self.subscripts.coefficients,
        )
        return self.add_or_transpose_modes(subscripts, dims)

    def permute_operands(self, perm: tuple[int, ...]) -> SegmentedTensorProduct:
        """Permute the operands of the descriptor."""
        # assert set(perm) == set(range(self.num_operands))  # removed for performance
        return dataclasses.replace(
            self,
            operands_and_subscripts=[self.operands_and_subscripts[i] for i in perm],
            paths=[path.permute_operands(perm) for path in self.paths],
        )

    def move_operand(self, operand: int, new_index: int) -> SegmentedTensorProduct:
        """Move an operand to a new index."""
        operand = _canonicalize_index("operand", operand, self.num_operands)
        if new_index < 0:
            new_index += self.num_operands
        perm = list(range(self.num_operands))
        del perm[operand]
        perm.insert(new_index, operand)
        return self.permute_operands(perm)

    def move_operand_first(self, operand: int) -> SegmentedTensorProduct:
        """Move an operand to the first position."""
        return self.move_operand(operand, 0)

    def move_operand_last(self, operand: int) -> SegmentedTensorProduct:
        """Move an operand to the last position."""
        return self.move_operand(operand, -1)

    def permute_segments(
        self, operand: int, perm: tuple[int, ...]
    ) -> SegmentedTensorProduct:
        """Permute the segments of an operand."""
        operand = _canonicalize_index("operand", operand, self.num_operands)
        new_operands = list(self.operands)
        new_operands[operand] = cue.SegmentedOperand(
            ndim=self.operands[operand].ndim,
            segments=[self.operands[operand][i] for i in perm],
        )
        new_paths = [
            Path(
                indices=tuple(
                    sid if oid != operand else perm.index(sid)
                    for oid, sid in enumerate(path.indices)
                ),
                coefficients=path.coefficients,
            )
            for path in self.paths
        ]
        return dataclasses.replace(
            self,
            operands_and_subscripts=list(zip(new_operands, self.subscripts.operands)),
            paths=new_paths,
        )

    def sort_paths(
        self, operands_ordering: Optional[Union[int, Sequence[int]]] = None
    ) -> SegmentedTensorProduct:
        """
        Sort the paths by their indices in lexicographical order.

        Args:
            operands_ordering (int or sequence of int, optional): The order in which to sort the paths.
                If an integer, sort by that operand.
                If a sequence, sort by the first operand in the sequence and then by the
                second operand if equal, etc.

        Returns:
            SegmentedTensorProduct: The sorted descriptor.
        """
        if operands_ordering is None:
            operands_ordering = range(self.num_operands)
        if isinstance(operands_ordering, int):
            operands_ordering = [operands_ordering]
        return dataclasses.replace(
            self,
            paths=sorted(
                self.paths,
                key=lambda path: tuple(path.indices[i] for i in operands_ordering),
            ),
        )

    def squeeze_modes(self, modes: Optional[str] = None) -> SegmentedTensorProduct:
        """
        Squeeze the descriptor by removing dimensions that are always 1.

        Args:
            modes (str, optional): The modes to squeeze. If None, squeeze all modes that are always 1.

        Returns:
            SegmentedTensorProduct: The squeezed descriptor.
        """
        to_remove = {m for m, dd in self.get_dimensions_dict().items() if dd == {1}}

        if modes is not None:
            modes: set[str] = set(modes)
            if not modes.issubset(to_remove):
                raise ValueError(
                    f"modes {modes} are not present in the dimensions that can be squeezed."
                )
            to_remove = to_remove & modes

        if not to_remove:
            return self

        d = SegmentedTensorProduct.from_subscripts(
            "".join(ch for ch in self.subscripts if ch not in to_remove)
        )

        def filter_shape(shape: tuple[int, ...], subscripts: str) -> tuple[int, ...]:
            return tuple(
                dim for dim, ch in zip(shape, subscripts) if ch not in to_remove
            )

        for oid, operand in enumerate(self.operands):
            d.add_segments(
                oid,
                [
                    filter_shape(segment, self.subscripts.operands[oid])
                    for segment in operand.segments
                ],
            )
        for path in self.paths:
            d.insert_path_(
                len(d.paths),
                Path(
                    indices=path.indices,
                    coefficients=np.reshape(
                        path.coefficients,
                        filter_shape(
                            path.coefficients.shape, self.coefficient_subscripts
                        ),
                    ),
                ),
            )
        logger.debug(f"Squeezed {self} to {d}")
        return d

    def split_mode(self, mode: str, size: int) -> SegmentedTensorProduct:
        """
        Split a mode into multiple modes of a given size.

        Args:
            mode (str): The mode to split. All its dimensions must be divisible by the size.
            size (int): The size of the new modes.

        Returns:
            SegmentedTensorProduct: The new descriptor.
        """
        if re.match(r"^[a-z]$", mode) is None:
            raise ValueError("expected a single lowercase letter.")

        if mode not in self.subscripts:
            return self

        for oid, ss in enumerate(self.subscripts.operands):
            if mode in ss and not ss.startswith(mode):
                raise ValueError(
                    f"mode {mode} is not the first mode in operand {oid} ({ss})."
                )

        if not all(dim % size == 0 for dim in self.get_dims(mode)):
            raise ValueError(f"mode {mode} is not divisible by {size}.")

        if (
            mode in self.coefficient_subscripts
            and not self.coefficient_subscripts.startswith(mode)
        ):
            return (
                self.add_or_transpose_modes(
                    Subscripts.from_operands(
                        self.subscripts.operands,
                        mode + self.coefficient_subscripts.replace(mode, ""),
                    )
                )
                .split_mode(mode, size)
                .add_or_transpose_modes(self.subscripts)
            )

        d = SegmentedTensorProduct.from_subscripts(self.subscripts)

        offsets_per_operand = []
        for oid, (operand, ss) in enumerate(self.operands_and_subscripts):
            if mode not in ss:
                for segment in operand:
                    d.add_segment(oid, segment)
                offsets_per_operand.append(None)
                continue

            assert ss.startswith(mode)

            offsets = []
            for segment in operand:
                offsets.append(d.operands[oid].num_segments)
                for _ in range(segment[0] // size):
                    d.add_segment(oid, (size,) + segment[1:])
            offsets_per_operand.append(offsets)

        for path in self.paths:
            num_subdivisions = self.get_path_dim(path, mode) // size

            for i in range(num_subdivisions):
                indices = []
                for oid in range(self.num_operands):
                    if offsets_per_operand[oid] is None:
                        indices.append(path.indices[oid])
                    else:
                        sid = path.indices[oid]
                        offset = offsets_per_operand[oid][sid]
                        indices.append(offset + i)
                coefficients = path.coefficients
                if self.coefficient_subscripts.startswith(mode):
                    coefficients = np.split(coefficients, num_subdivisions, axis=0)[i]
                d.insert_path_(
                    len(d.paths), Path(indices=indices, coefficients=coefficients)
                )

        logger.debug(f"Split {mode} in {self}: got {d}")
        return d

    def all_same_segment_shape(self) -> bool:
        """Check if all segments have the same shape."""
        for operand in self.operands:
            if not operand.all_same_segment_shape():
                return False
        return True

    def normalize_paths_for_operand(self, operand: int) -> SegmentedTensorProduct:
        """
        Normalize the paths for an operand.

        Args:
            operand (int): The index of the operand to normalize.

        Assuming that the input operand have unit variance, this method computes the
        variance of each segment in the selected operand and uniformly normalize the
        coefficients of the paths by the square root of the variance of the segment.
        This is useful to ensure that the output has unit variance.
        """
        operand = _canonicalize_index("operand", operand, self.num_operands)
        cum_variance = [0.0] * self.operands[operand].num_segments

        contracted_modes = {
            u
            for i, subscripts in enumerate(self.subscripts.operands_and_coefficients)
            #   ^^^^^^^-- (0, "uvw") (1, "iu") (2, "jv") (3, "kw") (4, "ijk")
            if i != operand  # e.g. discard (3, "kw")
            for u in subscripts  # u v w i u j v i j k
            if u not in self.subscripts.operands[operand]  # e.g. discard k and w
        }  # e.g. {u, v, i, j}

        for path in self.paths:
            dims = self.get_path_dimensions_dict(path)
            num = np.prod([dim for ch, dim in dims.items() if ch in contracted_modes])
            cum_variance[path.indices[operand]] += np.mean(path.coefficients**2) * num

        new_paths = []

        for path in self.paths:
            total_variance = cum_variance[path.indices[operand]]
            if total_variance > 0:
                c = path.coefficients / np.sqrt(total_variance)
            else:
                c = 0.0 * path.coefficients
            new_paths.append(Path(indices=path.indices, coefficients=c))

        d = dataclasses.replace(self, paths=new_paths)
        logger.debug(f"Normalized paths for operand {operand} in {self}: got {d}")
        return d

    def remove_zero_paths(self) -> SegmentedTensorProduct:
        """Remove paths with zero coefficients."""
        return dataclasses.replace(
            self,
            paths=[path for path in self.paths if not np.all(path.coefficients == 0)],
        )

    def fuse_paths_with_same_indices(self) -> SegmentedTensorProduct:
        """Fuse paths with the same indices."""
        paths = dict()
        for path in self.paths:
            if path.indices in paths:
                paths[path.indices] += path.coefficients
            else:
                paths[path.indices] = path.coefficients

        return dataclasses.replace(
            self,
            paths=[
                Path(indices=indices, coefficients=coefficients)
                for indices, coefficients in paths.items()
            ],
        )

    @functools.cache
    def consolidate_paths(self) -> SegmentedTensorProduct:
        """Consolidate the paths by merging duplicates and removing zeros."""
        # equivalent to self.fuse_paths_with_same_indices().remove_zero_paths().sort_paths()

        # TODO: use numpy when possible
        # if self.coefficients_are_stackable:
        #     indices = self.indices  # (num_paths, num_operands)
        #     coefficients = self.stacked_coefficients  # (num_paths, *coefficient_shape)

        paths = dict()
        for path in self.paths:
            if path.indices in paths:
                paths[path.indices] += path.coefficients
            else:
                paths[path.indices] = path.coefficients
        paths = [
            Path(indices=indices, coefficients=coefficients)
            for indices, coefficients in paths.items()
            if not np.all(coefficients == 0)
        ]
        paths = sorted(paths, key=lambda path: path.indices)
        return dataclasses.replace(self, paths=paths)

    def sort_indices_for_identical_operands(
        self, operands: Sequence[int]
    ) -> SegmentedTensorProduct:
        """Reduce the number of paths by sorting the indices for identical operands."""
        operands = tuple(sorted(set(operands)))
        if len(operands) < 2:
            return self

        return self._sort_indices_for_identical_operands(operands)

    @functools.cache
    def _sort_indices_for_identical_operands(self, operands: tuple[int, ...]):
        assert len({self.operands[oid].num_segments for oid in operands}) == 1

        non_trivial = any(
            m in self.coefficient_subscripts
            for i in operands
            for m in self.subscripts.operands[i]
        )
        if non_trivial:
            raise NotImplementedError(
                "missing code to handle sorting non scalar coefficients. Try to flatten the coefficients first."
            )

        def f(path: Path) -> Path:
            ii = path.indices
            aa = sorted([ii[oid] for oid in operands])
            return Path(
                indices=[
                    aa[operands.index(oid)] if oid in operands else ii[oid]
                    for oid in range(self.num_operands)
                ],
                coefficients=path.coefficients,
            )

        return dataclasses.replace(
            self, paths=[f(path) for path in self.paths]
        ).consolidate_paths()

    def symmetrize_operands(
        self, operands: Sequence[int], force: bool = False
    ) -> SegmentedTensorProduct:
        """Symmetrize the specified operands permuting the indices."""
        operands = tuple(sorted(set(operands)))
        if len(operands) < 2:
            return self

        return self._symmetrize_operands(operands, force)

    @functools.cache
    def _symmetrize_operands(self, operands: tuple[int, ...], force: bool):
        non_trivial = any(
            m in self.coefficient_subscripts
            for i in operands
            for m in self.subscripts.operands[i]
        )
        if non_trivial:
            raise NotImplementedError(
                "missing code to handle symmetrizing non scalar coefficients. Try to flatten the coefficients first."
            )

        permutations = list(itertools.permutations(range(len(operands))))

        def make_global_perm(perm: tuple[int, ...]) -> tuple[int, ...]:
            p = list(range(self.num_operands))
            for i, j in enumerate(perm):
                p[operands[i]] = operands[j]
            return tuple(p)

        if not force:
            # check if the tensor product is already symmetric
            symmetries: list[tuple[int, ...]] = self.symmetries()
            if all(make_global_perm(perm) in symmetries for perm in permutations):
                return self

        d = self.sort_indices_for_identical_operands(operands)

        paths = []
        for path in d.paths:
            indices = path.indices

            for perm in permutations:
                new_indices = list(indices)
                for i, oid in enumerate(operands):
                    new_indices[oid] = indices[operands[perm[i]]]
                paths.append(
                    Path(
                        indices=new_indices,
                        coefficients=path.coefficients / len(permutations),
                    )
                )

        d = dataclasses.replace(d, paths=paths)
        return d.consolidate_paths()

    def remove_empty_segments(self) -> SegmentedTensorProduct:
        """Remove empty segments."""

        if all(d > 0 for dd in self.get_dimensions_dict().values() for d in dd):
            return self

        def empty(D, oid, sid):
            return any(dim == 0 for dim in D.operands[oid][sid])

        D = self
        for oid in range(D.num_operands):
            perm = list(range(D.operands[oid].num_segments))
            for sid in range(D.operands[oid].num_segments):
                if empty(D, oid, sid):
                    perm.remove(sid)
                    perm.append(sid)
            D = D.permute_segments(oid, perm)

        D.set_paths(
            [
                path
                for path in D.paths
                if not any(empty(D, oid, sid) for oid, sid in enumerate(path.indices))
            ]
        )

        for oid in range(D.num_operands):
            operand = cue.SegmentedOperand(
                ndim=D.operands[oid].ndim,
                segments=[
                    segment
                    for sid, segment in enumerate(D.operands[oid].segments)
                    if not empty(D, oid, sid)
                ],
                _dims={
                    i: {d for d in dd if d > 0}
                    for i, dd in D.operands[oid]._dims.items()
                },
            )
            D.set_operand(oid, operand)

        return D

    def flatten_modes(
        self, modes: Sequence[str], *, skip_zeros=True, force=False
    ) -> SegmentedTensorProduct:
        """
        Remove the specified modes by subdividing segments and paths.

        Args:
            modes (Sequence of str): The modes to remove, they must precede the modes to keep in each operand.
            skip_zeros (bool, optional): Whether to skip paths with zero coefficients. Default is True.
            force (bool, optional): Whether to force the flattening by flattening extra necessary modes. Default is False.
        """
        if not all(len(ch) == 1 and ch.islower() for ch in modes):
            raise ValueError("expected lowercase single letter modes.")

        modes = {ch for ch in modes if ch in self.subscripts}
        modes = "".join(sorted(modes))

        if force:
            extra = ""
            for m in modes:
                for _, sm in self.operands_and_subscripts:
                    if m in sm:
                        extra += sm[: sm.index(m)]
            modes += extra

        modes = "".join(sorted(set(modes)))
        if len(modes) == 0:
            return self

        pattern = re.compile(rf"^([{modes}]*)([^{modes}]*)$")
        new_operands_and_subscripts = []
        offsets_per_operand = []
        rm_shape_per_operand = []
        rm_modes_per_operand = []
        for operand, subscripts in self.operands_and_subscripts:
            ma = pattern.match(subscripts)
            if ma is None:
                raise ValueError(
                    f"expected modes {modes} to be at the beginning of the segment subscripts."
                    f" Got {subscripts}."
                )

            rm_modes, new_subscripts = ma.groups()
            rm_modes_per_operand.append(rm_modes)

            n = len(rm_modes)  # number of axes to split

            new_segments = []

            off = 0
            offsets = []
            rm_shapes = []
            for segment in operand:
                offsets.append(off)
                num = math.prod(segment[:n])
                rm_shapes.append(segment[:n])
                new_segments += [segment[n:]] * num
                off += num
            offsets_per_operand.append(offsets)
            rm_shape_per_operand.append(rm_shapes)

            new_operands_and_subscripts.append(
                (
                    cue.SegmentedOperand(
                        ndim=len(new_subscripts), segments=new_segments
                    ),
                    new_subscripts,
                )
            )

        def ravel_multi_index(indices: tuple[int, ...], shape: tuple[int, ...]) -> int:
            if len(indices) == 0:
                # case not supported by some older numpy versions
                return 0
            return np.ravel_multi_index(indices, shape)

        def make_new_path(
            old_indices: tuple[int, ...],  # old segment indices (one per operand)
            sub_indices: dict[str, int],
            coefficients: np.ndarray,
        ) -> Path:
            return Path(
                [
                    offsets[sid]
                    + ravel_multi_index(
                        tuple(sub_indices[m] for m in rm_modes),
                        rm_shapes[sid],
                    )
                    for sid, offsets, rm_modes, rm_shapes in zip(
                        old_indices,
                        offsets_per_operand,
                        rm_modes_per_operand,
                        rm_shape_per_operand,
                    )
                ],
                coefficients,
            )

        new_paths = []
        if (
            skip_zeros
            and len(self.coefficient_subscripts) > 0
            and set(self.coefficient_subscripts).issubset(modes)
        ):
            non_coeff_modes = "".join(
                m for m in modes if m not in self.coefficient_subscripts
            )
            for path in self.paths:
                dims = self.get_path_dimensions_dict(path)
                idx = np.nonzero(path.coefficients)
                vals = path.coefficients[idx]

                for non_coeff_indices in np.ndindex(
                    tuple(dims[m] for m in non_coeff_modes)
                ):
                    d2 = dict(zip(non_coeff_modes, non_coeff_indices))
                    for coeff_indices, c in zip(zip(*idx), vals):
                        d1 = dict(zip(self.coefficient_subscripts, coeff_indices))

                        new_paths.append(
                            make_new_path(
                                path.indices,
                                d1 | d2,
                                c,
                            )
                        )
        else:
            for path in self.paths:
                dims = self.get_path_dimensions_dict(path)
                for index in np.ndindex(tuple(dims[m] for m in modes)):
                    c = path.coefficients[
                        tuple(
                            (index[modes.index(ch)] if ch in modes else slice(None))
                            for ch in self.coefficient_subscripts
                        )
                    ]
                    if skip_zeros and np.all(c == 0):
                        continue

                    new_paths.append(
                        make_new_path(path.indices, dict(zip(modes, index)), c)
                    )
        d = dataclasses.replace(
            self,
            operands_and_subscripts=new_operands_and_subscripts,
            paths=new_paths,
            coefficient_subscripts="".join(
                ch for ch in self.coefficient_subscripts if ch not in modes
            ),
        )
        logger.debug(f"Flattened {modes} in {self}: got {d}")
        return d

    def flatten_coefficient_modes(
        self, *, skip_zeros=True, force=False
    ) -> SegmentedTensorProduct:
        """
        Flatten the coefficients of the descriptor. Create new segments and paths to flatten the coefficients.

        Args:
            skip_zeros (bool, optional): Whether to skip paths with zero coefficients. Default is True.
            force (bool, optional): Whether to force the flattening by flattening extra necessary modes. Default is False.
        """
        return self.flatten_modes(
            self.coefficient_subscripts, skip_zeros=skip_zeros, force=force
        )

    def consolidate_modes(self, modes: Optional[str] = None) -> SegmentedTensorProduct:
        """
        Consolidate the descriptor by merging modes together.

        Args:
            modes (str, optional): The modes to consolidate. If None, consolidate all modes that are repeated.
        """
        if modes is None:
            # look for opportunities to consolidate
            for m in self.subscripts.modes():
                neighbors: set[str] = set()
                for _, subscripts in self.operands_and_subscripts:
                    if m in subscripts:
                        i = subscripts.index(m)
                        if i < len(subscripts) - 1:
                            neighbors.add(subscripts[i + 1])
                        else:
                            neighbors.add(".")
                if len(neighbors) != 1:
                    continue
                n = neighbors.pop()
                if not n.isalpha():
                    continue
                # Zuvw_Ziu_Zjv_Zkw+ijk

                ok = True
                for _, subscripts in self.operands_and_subscripts:
                    if m in subscripts:
                        if n not in subscripts:
                            ok = False
                            break
                    else:
                        if n in subscripts:
                            ok = False
                            break

                if not ok:
                    continue

                if m in self.subscripts.coefficients:
                    if n not in self.subscripts.coefficients:
                        continue
                else:
                    if n in self.subscripts.coefficients:
                        continue

                return self._consolidate_pair_of_modes(m, n).consolidate_modes()

            return self

        for _, subscripts in self.operands_and_subscripts:
            if modes not in subscripts:
                if any(ch in subscripts for ch in modes):
                    raise ValueError(
                        f"expected {modes} to be contiguous in the subscripts {subscripts}."
                    )

        d = self
        for n in modes[1:]:
            d = d._consolidate_pair_of_modes(modes[0], n)

        logger.debug(f"Consolidated {modes} in {self}: got {d}")
        return d

    def _consolidate_pair_of_modes(self, m: str, n: str) -> SegmentedTensorProduct:
        """keep m, absorb n in m"""
        d0 = self
        if m in d0.coefficient_subscripts:
            if n not in d0.coefficient_subscripts:
                raise ValueError(
                    f"expected both {m} and {n} in the coefficients subscripts. {d0.subscripts}"
                )
            tmp = d0.coefficient_subscripts.replace(m, "").replace(n, "")
            d0 = d0.add_or_transpose_modes(
                ",".join(d0.subscripts.operands) + f"+{m}{n}{tmp}"
            )

        d1 = SegmentedTensorProduct.from_subscripts(d0.subscripts.replace(m + n, m))

        for oid, (operand, subscripts) in enumerate(d0.operands_and_subscripts):
            for segment in operand:
                if m in subscripts:
                    i = subscripts.index(m)
                    segment = list(segment)
                    segment[i] *= segment[i + 1]
                    segment.pop(i + 1)
                    segment = tuple(segment)
                d1.add_segment(oid, segment)

        if m in d0.coefficient_subscripts:
            i = d0.coefficient_subscripts.index(m)
            for path in d0.paths:
                c = path.coefficients
                c = np.reshape(
                    c, c.shape[:i] + (c.shape[i] * c.shape[i + 1],) + c.shape[i + 2 :]
                )
                d1.insert_path_(
                    len(d1.paths), Path(indices=path.indices, coefficients=c)
                )
        else:
            d1.set_paths(d0.paths)

        return d1

    def round_coefficients_to_rational(
        self, max_denominator: int
    ) -> SegmentedTensorProduct:
        """
        Round the coefficients to the nearest ``p / q`` number with a given maximum denominator.

        Args:
            max_denominator (int): The maximum denominator, ``q < max_denominator``.
        """
        d = copy.deepcopy(self)
        d.set_paths(
            [
                Path(
                    indices=path.indices,
                    coefficients=round_to_rational(path.coefficients, max_denominator),
                )
                for path in d.paths
            ]
        )
        return d

    def round_coefficients_to_sqrt_rational(
        self, max_denominator: int
    ) -> SegmentedTensorProduct:
        """
        Round the coefficients to the nearest ``sqrt(p / q)`` number with a given maximum denominator.

        Args:
            max_denominator (int): The maximum denominator, ``q < max_denominator``.
        """
        d = copy.deepcopy(self)
        d.set_paths(
            [
                Path(
                    indices=path.indices,
                    coefficients=round_to_sqrt_rational(
                        path.coefficients, max_denominator
                    ),
                )
                for path in d.paths
            ]
        )
        return d

    def modify_coefficients(
        self, f: Callable[[np.ndarray], np.ndarray]
    ) -> SegmentedTensorProduct:
        """
        Modify the coefficients of the descriptor.

        Args:
            f (callable): The function to apply to the coefficients.
        """
        d = copy.deepcopy(self)
        d.set_paths(
            [
                Path(indices=path.indices, coefficients=f(path.coefficients))
                for path in d.paths
            ]
        )
        return d

    def __mul__(self, factor: float) -> SegmentedTensorProduct:
        """Amplify the path coefficients by a factor."""
        if factor == 1.0:
            return self
        return self.modify_coefficients(lambda c: factor * c)

    def __rmul__(self, factor: float) -> SegmentedTensorProduct:
        """Amplify the path coefficients by a factor."""
        return self * factor

    @property
    def slice_by_segment(self) -> _SegmentSlicer:
        """Return a slicer that allows slicing by segment index."""
        return _SegmentSlicer(self)

    @property
    def slice_by_size(self) -> _SizeSlicer:
        """Return a slicer that allows slicing by flat size/offset."""
        return _SizeSlicer(self)


class _SegmentSlicer:
    """Helper class for slicing SegmentedTensorProduct by segment index."""

    def __init__(self, stp: SegmentedTensorProduct):
        self.stp = stp

    def __getitem__(self, key) -> SegmentedTensorProduct:
        """
        Slice the SegmentedTensorProduct to get a subset by segment indices.

        Args:
            key: A tuple of slices for each operand.

        Returns:
            SegmentedTensorProduct: A new descriptor with sliced operands and filtered paths.

        Examples:
            >>> import cuequivariance as cue
            >>> stp = cue.SegmentedTensorProduct.from_subscripts("u,u")
            >>> stp.add_segment(0, (2,))
            0
            >>> stp.add_segment(0, (2,))
            1
            >>> stp.add_segment(1, (2,))
            0
            >>> stp.add_path(0, 0, c=1.0)
            0
            >>> sliced = stp.slice_by_segment[1:, :]
            >>> sliced.num_paths
            0
        """
        if not isinstance(key, tuple):
            key = (key,)

        if len(key) != self.stp.num_operands:
            raise ValueError(
                f"Expected a slice or int for each operand, got {len(key)} keys for {self.stp.num_operands} operands."
            )

        # Create new operands with sliced segments
        new_operands_and_subscripts = []
        segment_mappings = []  # Maps old segment indices to new ones

        for slice_obj, (operand, subscripts) in zip(
            key, self.stp.operands_and_subscripts
        ):
            if not isinstance(slice_obj, slice):
                raise TypeError(f"Invalid slice type: {type(slice_obj)}")

            # Apply slice to segments
            sliced_segments = operand.segments[slice_obj]
            new_operand = cue.SegmentedOperand(
                ndim=operand.ndim, segments=sliced_segments
            )

            # Create mapping from old to new segment indices
            old_indices = list(range(len(operand.segments)))[slice_obj]
            mapping = {old_idx: new_idx for new_idx, old_idx in enumerate(old_indices)}

            new_operands_and_subscripts.append((new_operand, subscripts))
            segment_mappings.append(mapping)

        # Filter and update paths
        new_paths = []
        for path in self.stp.paths:
            # Check if all referenced segments still exist
            new_indices = []
            valid_path = True

            for oid, old_segment_idx in enumerate(path.indices):
                if old_segment_idx in segment_mappings[oid]:
                    new_indices.append(segment_mappings[oid][old_segment_idx])
                else:
                    valid_path = False
                    break

            if valid_path:
                new_paths.append(
                    Path(indices=new_indices, coefficients=path.coefficients)
                )

        return SegmentedTensorProduct(
            operands_and_subscripts=new_operands_and_subscripts,
            coefficient_subscripts=self.stp.coefficient_subscripts,
            paths=new_paths,
        )


class _SizeSlicer:
    """Helper class for slicing SegmentedTensorProduct by flat size/offset."""

    def __init__(self, stp: SegmentedTensorProduct):
        self.stp = stp

    def __getitem__(self, key) -> SegmentedTensorProduct:
        """
        Slice the SegmentedTensorProduct to get a subset by flat size/offset.

        Args:
            key: A tuple of slices for each operand.

        Returns:
            SegmentedTensorProduct: A new descriptor with sliced operands and filtered paths.

        Examples:
            >>> import cuequivariance as cue
            >>> stp = cue.SegmentedTensorProduct.from_subscripts("u,u")
            >>> stp.add_segment(0, (2,))
            0
            >>> stp.add_segment(0, (2,))
            1
            >>> stp.add_segment(1, (2,))
            0
            >>> stp.add_path(0, 0, c=1.0)
            0
            >>> sliced = stp.slice_by_size[2:, :]
            >>> sliced.num_paths
            0
        """
        if not isinstance(key, tuple):
            key = (key,)

        if len(key) != self.stp.num_operands:
            raise ValueError(
                f"Expected a slice or int for each operand, got {len(key)} keys for {self.stp.num_operands} operands."
            )

        # Create new operands with sliced segments
        new_operands_and_subscripts = []
        segment_mappings = []  # Maps old segment indices to new ones

        for slice_obj, (operand, subscripts) in zip(
            key, self.stp.operands_and_subscripts
        ):
            if not isinstance(slice_obj, slice):
                raise TypeError(f"Invalid slice type: {type(slice_obj)}")

            # Use the operand's size-based slicing
            new_operand = operand.slice_by_size[slice_obj]

            # Create mapping from old to new segment indices
            # We need to find which segments are included in the size slice
            start, stop, step = slice_obj.indices(operand.size)
            if step != 1:
                raise ValueError("Step sizes other than 1 are not supported")

            mapping = {}
            offset = 0
            new_segment_idx = 0

            for old_segment_idx, segment in enumerate(operand.segments):
                segment_size = math.prod(segment)
                segment_start = offset
                segment_end = offset + segment_size

                # Check if this segment overlaps with [start, stop)
                if segment_start < stop and segment_end > start:
                    mapping[old_segment_idx] = new_segment_idx
                    new_segment_idx += 1

                offset += segment_size

                # If we've passed the stop point, we can break
                if offset >= stop:
                    break

            new_operands_and_subscripts.append((new_operand, subscripts))
            segment_mappings.append(mapping)

        # Filter and update paths
        new_paths = []
        for path in self.stp.paths:
            # Check if all referenced segments still exist
            new_indices = []
            valid_path = True

            for oid, old_segment_idx in enumerate(path.indices):
                if old_segment_idx in segment_mappings[oid]:
                    new_indices.append(segment_mappings[oid][old_segment_idx])
                else:
                    valid_path = False
                    break

            if valid_path:
                new_paths.append(
                    Path(indices=new_indices, coefficients=path.coefficients)
                )

        return SegmentedTensorProduct(
            operands_and_subscripts=new_operands_and_subscripts,
            coefficient_subscripts=self.stp.coefficient_subscripts,
            paths=new_paths,
        )


def _canonicalize_index(name: str, index: int, size: int) -> int:
    if not (-size <= index < size):
        raise ValueError(
            f"expected {name} to be between -{size} and {size - 1}, got {index}."
        )
    if index < 0:
        index += size
    return index



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from cuequivariance.group_theory.experimental.escn import escn_tp, escn_tp_compact

import cuequivariance as cue


def test_escn():
    escn_tp(
        cue.Irreps("O3", "32x0e + 32x1o"),
        cue.Irreps("O3", "32x0e + 32x1o"),
        m_max=2,
    )


def test_escn_compact():
    escn_tp_compact(
        cue.Irreps("SO3", "32x0 + 32x1"),
        cue.Irreps("SO3", "32x0 + 32x1"),
        m_max=2,
    )



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import itertools
import re
from typing import FrozenSet, Iterable, Optional, Sequence, Type, Union

import numpy as np

import cuequivariance as cue  # noqa: F401
from cuequivariance.group_theory.representations import Rep


# This class is inspired from https://github.com/lie-nn/lie-nn/blob/70adebce44e3197ee17f780585c6570d836fc2fe/lie_nn/_src/irrep.py
@dataclasses.dataclass(frozen=True)
class Irrep(Rep):
    r"""
    Subclass of :class:`Rep` for an irreducible representation of a Lie group.

    It extends the base class by adding:

    - A regular expression pattern for parsing the string representation.
    - The selection rule for the tensor product of two irreps.
    - An ordering relation for sorting the irreps.
    - A Clebsch-Gordan method for computing the Clebsch-Gordan coefficients.

    .. _custom-irreps:

    Simple example of defining custom irreps
    ----------------------------------------

    In some cases, you may want to define a custom set of irreducible representations of a group.
    Here is a simple example of how to define the irreps of the group :math:`Z_2`. For this we need to define a class that inherits from :class:`cue.Irrep <cuequivariance.Irrep>` and implement the required methods.

    .. jupyter-execute::

        from __future__ import annotations
        import re
        from typing import Iterator
        import numpy as np
        import cuequivariance as cue


        class Z2(cue.Irrep):
            odd: bool

            def __init__(rep: Z2, odd: bool):
                rep.odd = odd

            @classmethod
            def regexp_pattern(cls) -> re.Pattern:
                return re.compile(r"(odd|even)")

            @classmethod
            def from_string(cls, string: str) -> Z2:
                return cls(odd=string == "odd")

            def __repr__(rep: Z2) -> str:
                return "odd" if rep.odd else "even"

            def __mul__(rep1: Z2, rep2: Z2) -> Iterator[Z2]:
                return [Z2(odd=rep1.odd ^ rep2.odd)]

            @classmethod
            def clebsch_gordan(cls, rep1: Z2, rep2: Z2, rep3: Z2) -> np.ndarray:
                if rep3 in rep1 * rep2:
                    return np.array(
                        [[[[1]]]]
                    )  # (number_of_paths, rep1.dim, rep2.dim, rep3.dim)
                else:
                    return np.zeros((0, 1, 1, 1))

            @property
            def dim(rep: Z2) -> int:
                return 1

            def __lt__(rep1: Z2, rep2: Z2) -> bool:
                # False < True
                return rep1.odd < rep2.odd

            @classmethod
            def iterator(cls) -> Iterator[Z2]:
                for odd in [False, True]:
                    yield Z2(odd=odd)

            def discrete_generators(rep: Z2) -> np.ndarray:
                if rep.odd:
                    return -np.ones((1, 1, 1))  # (number_of_generators, rep.dim, rep.dim)
                else:
                    return np.ones((1, 1, 1))

            def continuous_generators(rep: Z2) -> np.ndarray:
                return np.zeros((0, rep.dim, rep.dim))  # (lie_dim, rep.dim, rep.dim)

            def algebra(self) -> np.ndarray:
                return np.zeros((0, 0, 0))  # (lie_dim, lie_dim, lie_dim)


        cue.Irreps(Z2, "13x odd + 6x even")

    .. rubric:: Methods to implement
    """

    @classmethod
    def regexp_pattern(cls) -> re.Pattern:
        """
        Regular expression pattern for parsing the string representation.
        """
        raise NotImplementedError

    @classmethod
    def from_string(cls, string: str) -> Irrep:
        """
        Create an instance from the string representation.
        """
        raise NotImplementedError

    @classmethod
    def _from(cls, *args) -> Irrep:
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, cls):
                return arg
            if isinstance(arg, str):
                return cls.from_string(arg)
            if isinstance(arg, Iterable):  # if isinstance(arg, tuple):
                return cls(
                    *iter(arg)
                )  # the iter is needed for compatibility with e3nn.o3.Irrep
        return cls(*args)

    def __repr__(rep: Irrep) -> str:
        raise NotImplementedError

    def __mul__(rep1: Irrep, rep2: Irrep) -> Iterable[Irrep]:
        """Selection rule for the tensor product of two irreps."""
        raise NotImplementedError

    def __lt__(rep1: Irrep, rep2: Irrep) -> bool:
        """
        This is required for sorting the irreps.

        - the dimension is the first criterion (ascending order)
        """
        if rep1 == rep2 or (rep1.dim != rep2.dim):
            return rep1.dim < rep2.dim
        raise NotImplementedError

    @classmethod
    def iterator(cls) -> Iterable[Irrep]:
        r"""
        Iterator over all irreps of the Lie group.

        - the first element is the trivial irrep
        - the elements respect the partial order defined by ``__lt__``
        """
        raise NotImplementedError

    @classmethod
    def clebsch_gordan(cls, rep1: Irrep, rep2: Irrep, rep3: Irrep) -> np.ndarray:
        """
        Clebsch-Gordan coefficients tensor.

        The shape is ``(number_of_paths, rep1.dim, rep2.dim, rep3.dim)`` and rep3 is the output irrep.

        See also:
            :func:`clebsch_gordan`.
        """
        raise NotImplementedError

    @classmethod
    def trivial(cls) -> Irrep:
        """
        Return the trivial irrep.

        Implemented by returning the first element of the iterator.
        """
        rep: Irrep = cls.iterator().__next__()
        assert rep.is_trivial(), "problem with the iterator"
        return rep


def clebsch_gordan(rep1: Irrep, rep2: Irrep, rep3: Irrep) -> np.ndarray:
    r"""
    Compute the Clebsch-Gordan coefficients.

    The Clebsch-Gordan coefficients are used to decompose the tensor product of two irreducible representations
    into a direct sum of irreducible representations. This method computes the Clebsch-Gordan coefficients
    for the given input representations and returns an array of shape ``(num_solutions, dim1, dim2, dim3)``,
    where num_solutions is the number of solutions, ``dim1`` is the dimension of ``rep1``, ``dim2`` is the
    dimension of ``rep2``, and ``dim3`` is the dimension of ``rep3``.

    The Clebsch-Gordan coefficients satisfy the following equation:

    .. math::

        C_{ljk} X^1_{li} + C_{ilk} X^2_{lj} = X^3_{kl} C_{ijl}

    Args:
        rep1 (Irrep): The first irreducible representation (input).
        rep2 (Irrep): The second irreducible representation (input).
        rep3 (Irrep): The third irreducible representation (output).

    Returns:
        np.ndarray: An array of shape ``(num_solutions, dim1, dim2, dim3)``.

    Examples:
        >>> rep1 = cue.SO3(1)
        >>> rep2 = cue.SO3(1)
        >>> rep3 = cue.SO3(2)
        >>> C = clebsch_gordan(rep1, rep2, rep3)
        >>> C.shape
        (1, 3, 3, 5)
        >>> C
        array([[[[ 0.  ...]]]])

        If there is no solution, the output is an empty array.

        >>> C = clebsch_gordan(cue.SO3(1), cue.SO3(1), cue.SO3(3))
        >>> C.shape
        (0, 3, 3, 7)
    """
    return rep1.clebsch_gordan(rep1, rep2, rep3)


def selection_rule_product(
    irs1: Union[Irrep, Sequence[Irrep], None],
    irs2: Union[Irrep, Sequence[Irrep], None],
) -> Optional[FrozenSet[Irrep]]:
    if irs1 is None or irs2 is None:
        return None

    if isinstance(irs1, Irrep):
        irs1 = [irs1]
    if isinstance(irs2, Irrep):
        irs2 = [irs2]
    irs1, irs2 = list(irs1), list(irs2)
    assert all(isinstance(x, Irrep) for x in irs1)
    assert all(isinstance(x, Irrep) for x in irs2)

    out = set()
    for ir1, ir2 in itertools.product(irs1, irs2):
        out = out.union(ir1 * ir2)
    return frozenset(out)


def selection_rule_power(
    irrep_class: Type[Irrep], irs: Union[Irrep, Sequence[Irrep]], n: int
) -> FrozenSet[Irrep]:
    if isinstance(irs, Irrep):
        irs = [irs]
    irs = list(irs)
    assert all(isinstance(x, Irrep) for x in irs)

    out = frozenset([irrep_class.trivial()])
    for _ in range(n):
        out = selection_rule_product(out, irs)
    return out



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import pytest

import cuequivariance as cue


def test_init():
    irreps = cue.Irreps("SU2", "1/2 + 16x2")
    array = np.zeros((128, irreps.dim))

    cue.NumpyIrrepsArray(irreps, array, cue.mul_ir)
    cue.NumpyIrrepsArray(irreps, array, cue.ir_mul)
    cue.NumpyIrrepsArray(irreps, array, "mul_ir")
    cue.NumpyIrrepsArray(irreps, array, "ir_mul")

    with pytest.raises(ValueError):
        cue.NumpyIrrepsArray(irreps, array, "invalid")
    with pytest.raises(ValueError):
        cue.NumpyIrrepsArray(irreps, np.zeros((128, 128, 128)), cue.mul_ir)


def test_sort():
    x = cue.NumpyIrrepsArray(cue.Irreps("SU2", "1/2 + 0 + 1"), np.arange(6), cue.mul_ir)
    y = x.sort()

    assert y.irreps == cue.Irreps("SU2", "0 + 1/2 + 1")
    np.testing.assert_array_equal(x.segments[0], y.segments[1])
    np.testing.assert_array_equal(x.segments[1], y.segments[0])
    np.testing.assert_array_equal(x.segments[2], y.segments[2])



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
from typing import Iterator

import numpy as np

import cuequivariance as cue


class O3_e3nn(cue.O3):
    def __mul__(rep1: "O3_e3nn", rep2: "O3_e3nn") -> Iterator["O3_e3nn"]:
        return [O3_e3nn(l=ir.l, p=ir.p) for ir in cue.O3.__mul__(rep1, rep2)]

    @classmethod
    def clebsch_gordan(
        cls, rep1: "O3_e3nn", rep2: "O3_e3nn", rep3: "O3_e3nn"
    ) -> np.ndarray:
        from e3nn import o3

        rep1, rep2, rep3 = cls._from(rep1), cls._from(rep2), cls._from(rep3)

        if rep1.p * rep2.p == rep3.p:
            return o3.wigner_3j(rep1.l, rep2.l, rep3.l).numpy()[None] * np.sqrt(
                rep3.dim
            )
        else:
            return np.zeros((0, rep1.dim, rep2.dim, rep3.dim))

    def __lt__(rep1: "O3_e3nn", rep2: "O3_e3nn") -> bool:
        rep2 = rep1._from(rep2)
        return (rep1.l, rep1.p) < (rep2.l, rep2.p)

    @classmethod
    def iterator(cls) -> Iterator["O3_e3nn"]:
        for ell in itertools.count(0):
            yield O3_e3nn(l=ell, p=1 * (-1) ** ell)
            yield O3_e3nn(l=ell, p=-1 * (-1) ** ell)



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from cuequivariance.group_theory.experimental.gatr import (
    gatr_geometric_product,
    gatr_linear,
    gatr_outer_product,
)


def test_geometric_product():
    gatr_geometric_product()


def test_outer_product():
    gatr_outer_product()


def test_linear():
    gatr_linear(32, 32)



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import math

from cuequivariance.segmented_polynomials.dimensions_dict import format_dimensions_dict


@dataclasses.dataclass(init=False, frozen=True)
class SegmentedOperand:
    """A segmented operand is a list of segment's shapes."""

    ndim: int
    segments: tuple[tuple[int, ...]]
    _dims: dict[int, set[int]]

    def __init__(
        self,
        segments: list[tuple[int, ...]] | None = None,
        *,
        ndim: int | None = None,
        _dims: dict[int, set[int]] | None = None,
    ):
        if segments is None:
            segments = []
        object.__setattr__(self, "segments", tuple(segments))

        if ndim is None:
            assert len(self.segments) > 0
            ndim = len(self.segments[0])
        object.__setattr__(self, "ndim", ndim)

        if _dims is None:
            _dims = dict()
            for segment in self.segments:
                for i, d in enumerate(segment):
                    _dims.setdefault(i, set()).add(d)
        else:
            _dims = _dims.copy()
        object.__setattr__(self, "_dims", _dims)

    @classmethod
    def empty_segments(cls, num_segments: int) -> SegmentedOperand:
        """Create an operand with ndim=0"""
        return cls(ndim=0, segments=[()] * num_segments, _dims=dict())

    @classmethod
    def stack(cls, operands: list[SegmentedOperand]) -> SegmentedOperand:
        """Stack a list of operands together."""
        assert len(operands) > 0
        ndim = operands[0].ndim
        assert all(ope.ndim == ndim for ope in operands)

        _dims = dict()
        for ope in operands:
            for i, d in ope.get_dimensions_dict().items():
                _dims.setdefault(i, set()).update(d)

        return cls(
            ndim=ndim,
            segments=sum([list(ope.segments) for ope in operands], []),
            _dims=_dims,
        )

    def copy(self) -> SegmentedOperand:
        """Copy the operand."""
        return SegmentedOperand(
            ndim=self.ndim,
            segments=self.segments,
            _dims=self._dims,
        )

    def assert_valid(self):
        """Assert that the operand is valid."""
        for segment in self.segments:
            if len(segment) != self.ndim:
                raise ValueError(
                    f"segment {segment} has {len(segment)} dimensions, expected {self.ndim}."
                )

            if not all(isinstance(dim, int) and dim > 0 for dim in segment):
                raise ValueError(f"segment {segment} is not valid.")

            for i, d in enumerate(segment):
                if d not in self.get_dims(i):
                    raise ValueError(
                        f"dimension {d} not in {i} dimensions {self.get_dims(i)}."
                    )

    def insert_segment(self, index: int, segment: tuple[int, ...]):
        """Insert a segment at a given index."""
        if len(segment) != self.ndim:
            raise ValueError(
                f"segment has {len(segment)} dimensions, expected {self.ndim}."
            )

        if index < 0:
            index = len(self.segments) + index

        if index < 0 or index > len(self.segments):
            raise ValueError(
                f"index {index} is out of bounds for segments {self.segments}."
            )

        segment = tuple(int(d) for d in segment)
        object.__setattr__(
            self,
            "segments",
            self.segments[:index] + (segment,) + self.segments[index:],
        )

        for i, d in enumerate(segment):
            self._dims.setdefault(i, set()).add(d)

    def add_segment(self, segment: tuple[int, ...]) -> int:
        """Add a segment to the operand."""
        self.insert_segment(len(self.segments), segment)
        return len(self.segments) - 1

    def __hash__(self) -> int:
        return hash((self.ndim, self.segments))

    def __eq__(self, other: SegmentedOperand) -> bool:
        assert isinstance(other, SegmentedOperand)
        return self.ndim == other.ndim and self.segments == other.segments

    def __lt__(self, other: SegmentedOperand) -> bool:
        assert isinstance(other, SegmentedOperand)
        return (self.ndim, self.segments) < (other.ndim, other.segments)

    def __repr__(self) -> str:
        dims = format_dimensions_dict(self.get_dimensions_dict())
        return f"Operand(ndim={self.ndim} num_segments={self.num_segments} dims={dims})"

    def __getitem__(self, index: int) -> tuple[int, ...]:
        return self.segments[index]

    def __len__(self) -> int:
        return self.num_segments

    def __iter__(self):
        return iter(self.segments)

    @property
    def num_segments(self) -> int:
        """The number of segments in the operand."""
        return len(self.segments)

    @property
    def size(self) -> int:
        """The total size of the operand."""
        if self.all_same_segment_shape():
            return self.num_segments * self.segment_size

        return sum(math.prod(segment) for segment in self.segments)

    def segment_slices(self) -> list[slice]:
        """Return slice object for each segment."""
        offset = 0
        slices = []
        for segment in self.segments:
            slices.append(slice(offset, offset + math.prod(segment)))
            offset += math.prod(segment)
        return slices

    def get_dimensions_dict(self) -> dict[int, set[int]]:
        """Return a dictionary of dimensions for each channel."""
        return self._dims.copy()

    def get_dims(self, i: int) -> set[int]:
        """Return the dimensions for a given channel."""
        return self._dims.get(i, set()).copy()

    def all_same_segment_shape(self) -> bool:
        """Check if all segments have the same shape. Returns False if there are no segments."""
        return all(len(dd) == 1 for dd in self._dims.values()) and self.num_segments > 0

    @property
    def segment_shape(self) -> tuple[int, ...]:
        """The shape of the segments if they are all the same."""
        if not self.all_same_segment_shape():
            raise ValueError("Segments do not have the same shape.")
        return self.segments[0]

    @property
    def segment_size(self) -> int:
        """The size of the segments if they are all the same."""
        if not self.all_same_segment_shape():
            raise ValueError("Segments do not have the same shape.")
        return math.prod(self.segments[0])

    def __add__(self, other: SegmentedOperand) -> SegmentedOperand:
        if self.ndim != other.ndim:
            raise ValueError("ndim do not match.")
        return SegmentedOperand(
            ndim=self.ndim,
            segments=self.segments + other.segments,
            _dims={i: self.get_dims(i) | other.get_dims(i) for i in range(self.ndim)},
        )

    @property
    def slice_by_segment(self) -> _SegmentSlicer:
        """Return a slicer that allows slicing by segment index."""
        return _SegmentSlicer(self)

    @property
    def slice_by_size(self) -> _SizeSlicer:
        """Return a slicer that allows slicing by flat size/offset."""
        return _SizeSlicer(self)


class _SegmentSlicer:
    """Helper class for slicing by segment index."""

    def __init__(self, operand: SegmentedOperand):
        self.operand = operand

    def __getitem__(self, key: slice) -> SegmentedOperand:
        if not isinstance(key, slice):
            raise TypeError(f"Only slice objects are supported, got {type(key)}")
        return SegmentedOperand(
            ndim=self.operand.ndim, segments=self.operand.segments[key]
        )


class _SizeSlicer:
    """Helper class for slicing by size/offset."""

    def __init__(self, operand: SegmentedOperand):
        self.operand = operand

    def __getitem__(self, key: slice) -> SegmentedOperand:
        if not isinstance(key, slice):
            raise TypeError(f"Only slice objects are supported, got {type(key)}")

        # Handle slice
        start, stop, step = key.indices(self.operand.size)

        if step != 1:
            raise ValueError("Step sizes other than 1 are not supported")

        # Find segments that overlap with [start, stop)
        segments = []
        offset = 0

        for segment in self.operand.segments:
            segment_size = math.prod(segment)
            segment_start = offset
            segment_end = offset + segment_size

            # Check if this segment overlaps with [start, stop)
            if segment_start < stop and segment_end > start:
                segments.append(segment)

            offset += segment_size

            # If we've passed the stop point, we can break
            if offset >= stop:
                break

        return SegmentedOperand(ndim=self.operand.ndim, segments=segments)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np

import cuequivariance as cue


def make_simple_stp() -> cue.SegmentedTensorProduct:
    d = cue.SegmentedTensorProduct.empty_segments([2, 2, 2])
    d.add_path(0, 0, 0, c=1.0)
    d.add_path(1, 1, 1, c=-2.0)
    return d


def make_simple_dot_product_stp() -> cue.SegmentedTensorProduct:
    d = cue.SegmentedTensorProduct.from_subscripts("i,j,k+ijk")
    i0 = d.add_segment(0, (3,))
    i1 = d.add_segment(1, (3,))
    i2 = d.add_segment(2, (1,))
    d.add_path(i0, i1, i2, c=np.eye(3).reshape(3, 3, 1))
    return d


def test_init_segmented_polynomial():
    """Test initialization of SegmentedPolynomial."""
    stp = make_simple_stp()
    poly = cue.SegmentedPolynomial.eval_last_operand(stp)

    assert poly.num_inputs == 2 and poly.num_outputs == 1 and poly.num_operands == 3
    assert len(poly.operations) == 1
    assert poly.operations[0] == (cue.Operation((0, 1, 2)), stp)


def test_polynomial_equality():
    """Test equality comparison of polynomials."""
    stp1 = make_simple_stp()
    stp2 = make_simple_stp()

    poly1 = cue.SegmentedPolynomial.eval_last_operand(stp1)
    poly2 = cue.SegmentedPolynomial.eval_last_operand(stp2)
    poly3 = cue.SegmentedPolynomial.eval_last_operand(2 * stp2)

    assert poly1 == poly2 and poly1 != poly3 and poly1 < poly3


def test_call_function():
    """Test calling the polynomial as a function."""
    stp = make_simple_dot_product_stp()
    poly = cue.SegmentedPolynomial.eval_last_operand(stp)

    a = np.array([1.0, 2.0, 3.0])
    b = np.array([4.0, 5.0, 6.0])

    [result] = poly(a, b)
    assert np.allclose(result, np.array([a.dot(b)]))


def test_buffer_properties():
    """Test properties related to buffer sizes and usage."""
    stp1 = make_simple_stp()
    op1 = cue.Operation((0, 1, 2))

    stp2 = cue.SegmentedTensorProduct.empty_segments([2, 1])
    stp2.add_path(0, 0, c=1.0)
    op2 = cue.Operation((0, 3))

    poly = cue.SegmentedPolynomial(
        [
            cue.SegmentedOperand.empty_segments(2),
            cue.SegmentedOperand.empty_segments(2),
        ],
        [
            cue.SegmentedOperand.empty_segments(2),
            cue.SegmentedOperand.empty_segments(1),
        ],
        [(op1, stp1), (op2, stp2)],
    )

    assert [ope.size for ope in poly.operands] == [2, 2, 2, 1]
    assert poly.used_operands() == [True, True, True, True]


def test_remove_unused_buffers():
    """Test removing unused buffers from the polynomial."""
    stp = make_simple_stp()
    op = cue.Operation((0, 2, 3))  # Buffer 1 is not used

    poly = cue.SegmentedPolynomial(
        [
            cue.SegmentedOperand.empty_segments(2),
            cue.SegmentedOperand.empty_segments(2),
            cue.SegmentedOperand.empty_segments(2),
        ],
        [cue.SegmentedOperand.empty_segments(2)],
        [(op, stp)],
    )

    assert poly.used_operands() == [True, False, True, True]

    cleaned = poly.filter_drop_unsued_operands()
    assert cleaned.num_inputs == 2 and cleaned.num_outputs == 1
    assert cleaned.used_operands() == [True, True, True]


def test_consolidate():
    """Test consolidating tensor products."""
    stp1 = make_simple_stp()
    stp2 = make_simple_stp()
    op = cue.Operation((0, 1, 2))

    poly = cue.SegmentedPolynomial(
        [
            cue.SegmentedOperand.empty_segments(2),
            cue.SegmentedOperand.empty_segments(2),
        ],
        [cue.SegmentedOperand.empty_segments(2)],
        [(op, stp1), (op, stp2)],
    )

    consolidated = poly.consolidate()
    assert len(consolidated.operations) == 1
    assert len(consolidated.operations[0][1].paths) == 2
    assert consolidated.operations[0][1].paths[0].coefficients == 2.0
    assert consolidated.operations[0][1].paths[1].coefficients == -4.0


def test_stack():
    """Test stacking polynomials."""
    stp1 = cue.SegmentedTensorProduct.empty_segments([2, 2, 1])
    stp1.add_path(0, 0, 0, c=1.0)
    poly1 = cue.SegmentedPolynomial.eval_last_operand(stp1)

    stp2 = cue.SegmentedTensorProduct.empty_segments([2, 2, 1])
    stp2.add_path(0, 0, 0, c=2.0)
    poly2 = cue.SegmentedPolynomial.eval_last_operand(stp2)

    stacked = cue.SegmentedPolynomial.stack([poly1, poly2], [False, False, True])
    assert stacked.num_inputs == 2 and stacked.num_outputs == 1
    assert [ope.size for ope in stacked.operands] == [2, 2, 2]


def test_flops_and_memory():
    """Test computation of FLOPS and memory usage."""
    stp = make_simple_stp()
    op = cue.Operation((0, 1, 2))
    poly = cue.SegmentedPolynomial(
        [
            cue.SegmentedOperand.empty_segments(2),
            cue.SegmentedOperand.empty_segments(2),
        ],
        [cue.SegmentedOperand.empty_segments(2)],
        [(op, stp)],
    )

    assert poly.flop(batch_size=100) > 0
    assert poly.memory([100, 100, 100]) == 100 * (2 + 2 + 2)


def test_jvp():
    """Test Jacobian-vector product computation."""
    stp = make_simple_dot_product_stp()
    poly = cue.SegmentedPolynomial.eval_last_operand(stp)

    x = np.array([1.0, 2.0, 3.0])
    y = np.array([4.0, 5.0, 6.0])
    x_tangent = np.array([0.1, 0.2, 0.3])
    y_tangent = np.array([0.4, 0.5, 0.6])

    # Test with both inputs having tangents
    jvp_poly, map = poly.jvp([True, True])
    assert map(([0, 1], [2])) == ([0, 1, 0, 1], [2])

    jvp_result = jvp_poly(x, y, x_tangent, y_tangent)
    expected_jvp = np.array([y.dot(x_tangent) + x.dot(y_tangent)])
    assert np.allclose(jvp_result[0], expected_jvp)

    # Test with only x having tangent
    jvp_x_only, _ = poly.jvp([True, False])
    x_only_result = jvp_x_only(x, y, x_tangent)
    assert np.allclose(x_only_result[0], np.array([y.dot(x_tangent)]))

    # Test with only y having tangent
    jvp_y_only, _ = poly.jvp([False, True])
    y_only_result = jvp_y_only(x, y, y_tangent)
    assert np.allclose(y_only_result[0], np.array([x.dot(y_tangent)]))


def test_transpose_linear():
    """Test transposing a linear polynomial."""
    stp = make_simple_dot_product_stp()
    poly = cue.SegmentedPolynomial.eval_last_operand(stp)

    x = np.array([1.0, 2.0, 3.0])
    y = np.array([4.0, 5.0, 6.0])
    cotangent = np.array([2.0])

    # Test transpose w.r.t. x
    transpose_x, _ = poly.transpose(
        is_undefined_primal=[True, False], has_cotangent=[True]
    )
    x_result = transpose_x(y, cotangent)
    assert np.allclose(x_result[0], y * cotangent[0])

    # Test transpose w.r.t. y
    transpose_y, _ = poly.transpose(
        is_undefined_primal=[False, True], has_cotangent=[True]
    )
    y_result = transpose_y(x, cotangent)
    assert np.allclose(y_result[0], x * cotangent[0])


def test_transpose_nonlinear():
    """Test transposing a non-linear polynomial raises an error."""
    stp = make_simple_stp()
    op = cue.Operation((0, 0, 1))  # Using same buffer twice (x^2)
    poly = cue.SegmentedPolynomial(
        [cue.SegmentedOperand.empty_segments(2)],
        [cue.SegmentedOperand.empty_segments(2)],
        [(op, stp)],
    )

    with np.testing.assert_raises(ValueError):
        poly.transpose(is_undefined_primal=[True], has_cotangent=[True])


def test_backward():
    """Test the backward method for gradient computation."""
    stp = make_simple_dot_product_stp()
    poly = cue.SegmentedPolynomial.eval_last_operand(stp)

    x = np.array([1.0, 2.0, 3.0])
    y = np.array([4.0, 5.0, 6.0])
    cotangent = np.array([2.0])

    # Test backward for both inputs
    backward_both, _ = poly.backward(
        requires_gradient=[True, True], has_cotangent=[True]
    )
    grad_x, grad_y = backward_both(x, y, cotangent)
    assert np.allclose(grad_x, y * cotangent[0]) and np.allclose(
        grad_y, x * cotangent[0]
    )

    # Test backward for x only
    backward_x, _ = poly.backward(requires_gradient=[True, False], has_cotangent=[True])
    [grad_x_only] = backward_x(x, y, cotangent)
    assert np.allclose(grad_x_only, y * cotangent[0])

    # Test backward for y only
    backward_y, _ = poly.backward(requires_gradient=[False, True], has_cotangent=[True])
    [grad_y_only] = backward_y(x, y, cotangent)
    assert np.allclose(grad_y_only, x * cotangent[0])

    # Test with zero cotangent
    grad_x_zero, grad_y_zero = backward_both(x, y, np.array([0.0]))
    assert np.allclose(grad_x_zero, np.zeros_like(x)) and np.allclose(
        grad_y_zero, np.zeros_like(y)
    )


def test_symmetrize_identical_operands():
    """Test symmetrization and unsymmetrization of polynomials with identical operands."""
    stp = cue.SegmentedTensorProduct.empty_segments([2, 2, 1])
    stp.add_path(0, 1, 0, c=1.0)

    op = cue.Operation((0, 0, 1))
    poly = cue.SegmentedPolynomial(
        [cue.SegmentedOperand.empty_segments(2)],
        [cue.SegmentedOperand.empty_segments(1)],
        [(op, stp)],
    )

    sym_poly = poly.symmetrize_for_identical_operands()
    [(_, sym_stp)] = sym_poly.operations

    assert len(sym_stp.paths) == 2
    assert sym_stp.paths[0].coefficients == sym_stp.paths[1].coefficients == 0.5
    assert sym_stp.paths[0].indices == (0, 1, 0) and sym_stp.paths[1].indices == (
        1,
        0,
        0,
    )

    unsym_poly = sym_poly.unsymmetrize_for_identical_operands()
    [(_, unsym_stp)] = unsym_poly.operations
    assert len(unsym_stp.paths) == 1 and unsym_stp.paths[0].coefficients == 1.0

    x = np.array([1.0, 2.0])
    assert np.allclose(poly(x)[0], sym_poly(x)[0])


def test_stack_tensor_products():
    """Test stacking tensor products together."""
    stp1 = cue.SegmentedTensorProduct.empty_segments([2, 2, 2])
    stp1.add_path(0, 0, 0, c=1.0)

    stp2 = cue.SegmentedTensorProduct.empty_segments([2, 2, 1])
    stp2.add_path(0, 1, 0, c=2.0)

    in1 = cue.SegmentedOperand.empty_segments(2)
    in2 = cue.SegmentedOperand.empty_segments(2)
    poly = cue.SegmentedPolynomial.stack_tensor_products(
        [in1, in2], [None], [([0, 1, 2], stp1), ([0, 1, 2], stp2)]
    )

    assert poly.num_inputs == 2 and poly.num_outputs == 1 and poly.num_operands == 3
    assert len(poly.operations) == 1
    assert poly.outputs[0].num_segments == 3


def test_concatenate():
    """Test concatenating segmented polynomials."""
    stp1 = cue.SegmentedTensorProduct.empty_segments([2, 1])
    stp1.add_path(0, 0, c=1.0)
    poly1 = cue.SegmentedPolynomial(
        [cue.SegmentedOperand.empty_segments(2)],
        [cue.SegmentedOperand.empty_segments(1)],
        [(cue.Operation((0, 1)), stp1)],
    )

    stp2 = cue.SegmentedTensorProduct.empty_segments([2, 1])
    stp2.add_path(0, 0, c=2.0)
    poly2 = cue.SegmentedPolynomial(
        [cue.SegmentedOperand.empty_segments(2)],
        [cue.SegmentedOperand.empty_segments(1)],
        [(cue.Operation((0, 1)), stp2)],
    )

    [in1] = poly1.inputs
    [out1] = poly1.outputs
    [out2] = poly2.outputs

    combined = cue.SegmentedPolynomial.concatenate(
        [in1], [out1, out2], [(poly1, [0, 1, None]), (poly2, [0, None, 1])]
    )

    assert combined.num_inputs == 1 and combined.num_outputs == 2
    assert len(combined.operations) == 2

    x = np.array([1.0, 2.0])
    y1, y2 = combined(x)
    assert np.isclose(y1[0], x[0]) and np.isclose(y2[0], 2.0 * x[0])


def test_filter_keep_outputs():
    """Test filtering to keep only selected outputs."""
    in_op = cue.SegmentedOperand.empty_segments(3)
    out1 = cue.SegmentedOperand.empty_segments(2)
    out2 = cue.SegmentedOperand.empty_segments(1)

    stp1 = cue.SegmentedTensorProduct.empty_segments([3, 2])
    stp1.add_path(0, 0, c=1.0)
    stp1.add_path(1, 1, c=1.0)
    op1 = cue.Operation((0, 1))

    stp2 = cue.SegmentedTensorProduct.empty_segments([3, 1])
    stp2.add_path(2, 0, c=2.0)
    op2 = cue.Operation((0, 2))

    poly = cue.SegmentedPolynomial([in_op], [out1, out2], [(op1, stp1), (op2, stp2)])
    filtered = poly.filter_keep_outputs([True, False])

    assert filtered.num_inputs == 1 and filtered.num_outputs == 1
    assert len(filtered.operations) == 1

    test_input = np.array([1.0, 2.0, 3.0])
    [result] = filtered(test_input)
    assert (
        result.shape == (2,)
        and np.isclose(result[0], test_input[0])
        and np.isclose(result[1], test_input[1])
    )


def test_fuse_stps():
    """Test fusing segmented tensor products with identical operations."""
    input_op = cue.SegmentedOperand.empty_segments(2)
    output_op = cue.SegmentedOperand.empty_segments(1)

    stp1 = cue.SegmentedTensorProduct.empty_segments([2, 1])
    stp1.add_path(0, 0, c=1.0)

    stp2 = cue.SegmentedTensorProduct.empty_segments([2, 1])
    stp2.add_path(0, 0, c=2.0)

    op = cue.Operation((0, 1))
    poly = cue.SegmentedPolynomial([input_op], [output_op], [(op, stp1), (op, stp2)])

    assert len(poly.operations) == 2

    fused = poly.fuse_stps()
    assert len(fused.operations) == 1

    fused_op, fused_stp = fused.operations[0]
    assert fused_stp.paths[0].coefficients == 3.0

    test_input = np.array([1.0, 2.0])
    assert np.allclose(poly(test_input), fused(test_input))


def test_compute_only():
    """Test creating a polynomial that only computes selected outputs."""
    input_op = cue.SegmentedOperand.empty_segments(2)
    output1 = cue.SegmentedOperand.empty_segments(1)
    output2 = cue.SegmentedOperand.empty_segments(1)

    stp1 = cue.SegmentedTensorProduct.empty_segments([2, 1])
    stp1.add_path(0, 0, c=1.0)
    op1 = cue.Operation((0, 1))

    stp2 = cue.SegmentedTensorProduct.empty_segments([2, 1])
    stp2.add_path(1, 0, c=2.0)
    op2 = cue.Operation((0, 2))

    poly = cue.SegmentedPolynomial(
        [input_op], [output1, output2], [(op1, stp1), (op2, stp2)]
    )
    filtered = poly.compute_only([False, True])

    assert (
        filtered.num_inputs == poly.num_inputs
        and filtered.num_outputs == poly.num_outputs
    )
    assert len(filtered.operations) == 1 and filtered.operations[0][0] == op2

    test_input = np.array([1.0, 2.0])
    full_output = poly(test_input)
    filtered_output = filtered(test_input)

    assert np.all(filtered_output[0] == 0)
    assert np.array_equal(filtered_output[1], full_output[1])


def test_permute_inputs_and_outputs():
    """Test permuting inputs and outputs of a SegmentedPolynomial."""
    # Create 3 inputs and 3 outputs
    inputs = [
        cue.SegmentedOperand.empty_segments(2),  # in1
        cue.SegmentedOperand.empty_segments(3),  # in2
        cue.SegmentedOperand.empty_segments(4),  # in3
    ]
    outputs = [
        cue.SegmentedOperand.empty_segments(1),  # out1
        cue.SegmentedOperand.empty_segments(1),  # out2
        cue.SegmentedOperand.empty_segments(1),  # out3
    ]

    # Create operations: each input i maps to output i
    operations = []
    for i in range(3):
        stp = cue.SegmentedTensorProduct.empty_segments([inputs[i].size, 1])
        stp.add_path(i, 0, c=(i + 1.0))
        operations.append((cue.Operation((i, i + 3)), stp))

    poly = cue.SegmentedPolynomial(inputs, outputs, operations)

    # Test data
    x1 = np.array([1.0, 2.0])
    x2 = np.array([3.0, 4.0, 5.0])
    x3 = np.array([6.0, 7.0, 8.0, 9.0])

    # Original results
    y1, y2, y3 = poly(x1, x2, x3)

    # Test input permutation [2,0,1]
    perm_in = poly.permute_inputs([2, 0, 1])
    y1_in, y2_in, y3_in = perm_in(x3, x1, x2)  # Inputs permuted

    # Structure checks
    assert [op.size for op in perm_in.inputs] == [4, 2, 3]
    assert perm_in.operations[0][0].buffers[0] == 1  # Buffer references updated
    assert perm_in.operations[1][0].buffers[0] == 2
    assert perm_in.operations[2][0].buffers[0] == 0

    # Output checks - results should match
    assert np.isclose(y1_in, y1)
    assert np.isclose(y2_in, y2)
    assert np.isclose(y3_in, y3)

    # Test output permutation [1,2,0]
    perm_out = poly.permute_outputs([1, 2, 0])
    y2_out, y3_out, y1_out = perm_out(x1, x2, x3)  # Outputs permuted

    # Structure checks
    assert [op.size for op in perm_out.outputs] == [1, 1, 1]
    assert perm_out.operations[0][0].buffers[1] == 3  # Buffer indices preserved
    assert perm_out.operations[1][0].buffers[1] == 4
    assert perm_out.operations[2][0].buffers[1] == 5

    # Output checks - results should match but in different order
    assert np.isclose(y1_out, y1)
    assert np.isclose(y2_out, y2)
    assert np.isclose(y3_out, y3)


def test_split_operand_by():
    """Test splitting operands by size and segment."""

    # ========================================
    # Example 1: Linear polynomial with contractions
    # ========================================

    # Create a polynomial with 4 segments in first input, 3 in second input
    stp = cue.SegmentedTensorProduct.from_subscripts("u,u,")

    # First input: 4 segments of shape (16,) each
    stp.add_segment(0, (16,))  # segment 0
    stp.add_segment(0, (16,))  # segment 1
    stp.add_segment(0, (16,))  # segment 2
    stp.add_segment(0, (16,))  # segment 3

    # Second input: 3 segments of shape (16,) each
    stp.add_segment(1, (16,))  # segment 0
    stp.add_segment(1, (16,))  # segment 1
    stp.add_segment(1, (16,))  # segment 2

    # Output: 1 segment (scalar)
    stp.add_segment(2, ())

    # Add paths (contractions between segments)
    stp.add_path(0, 0, 0, c=1.0)  # input0[seg0] * input1[seg0]
    stp.add_path(1, 0, 0, c=2.0)  # input0[seg1] * input1[seg0]
    stp.add_path(2, 1, 0, c=3.0)  # input0[seg2] * input1[seg1]
    stp.add_path(3, 2, 0, c=4.0)  # input0[seg3] * input1[seg2]

    poly = cue.SegmentedPolynomial.eval_last_operand(stp)

    # Test data
    x = np.random.rand(64)  # 4 segments × 16 = 64
    y = np.random.rand(48)  # 3 segments × 16 = 48

    # Original evaluation
    [result_original] = poly(x, y)

    # ========================================
    # Test 1a: Split by size
    # ========================================

    # Split first input at every 16 elements (segment boundaries)
    poly_split_size = poly.split_operand_by_size(0, [0, 16, 32, 48, 64])

    # Now we have 5 inputs: 4 split parts + original second input
    assert poly_split_size.num_inputs == 5

    # Evaluate with split inputs
    [result_split_size] = poly_split_size(
        x[0:16],  # first segment
        x[16:32],  # second segment
        x[32:48],  # third segment
        x[48:64],  # fourth segment
        y,  # second input unchanged
    )

    # Results should match
    np.testing.assert_allclose(result_original, result_split_size)

    # ========================================
    # Test 1b: Split by segment
    # ========================================

    # Split first input into two groups: segments [0,1] and segments [2,3]
    poly_split_seg = poly.split_operand_by_segment(0, [0, 2, 4])

    # Now we have 3 inputs: 2 split parts + original second input
    assert poly_split_seg.num_inputs == 3

    # Evaluate with split inputs
    [result_split_seg] = poly_split_seg(
        x[0:32],  # segments 0 and 1 (2 × 16 = 32)
        x[32:64],  # segments 2 and 3 (2 × 16 = 32)
        y,  # second input unchanged
    )

    # Results should match
    np.testing.assert_allclose(result_original, result_split_seg)

    # ========================================
    # Example 2: Quadratic polynomial (x² terms)
    # ========================================

    # Create a polynomial that computes quadratic terms
    stp_quad = cue.SegmentedTensorProduct.from_subscripts("u,u,")

    # Input has 2 segments of shape (8,) each
    stp_quad.add_segment(0, (8,))  # segment 0
    stp_quad.add_segment(0, (8,))  # segment 1

    # For quadratic terms, we need the same segments again
    stp_quad.add_segment(1, (8,))  # segment 0
    stp_quad.add_segment(1, (8,))  # segment 1

    # Output: 1 segment (scalar)
    stp_quad.add_segment(2, ())

    # Add paths for quadratic terms
    stp_quad.add_path(0, 0, 0, c=1.0)  # x[0] * x[0]
    stp_quad.add_path(1, 1, 0, c=2.0)  # x[1] * x[1]
    stp_quad.add_path(0, 1, 0, c=3.0)  # x[0] * x[1]
    stp_quad.add_path(1, 0, 0, c=4.0)  # x[1] * x[0]

    # Create polynomial where both inputs are the same operand
    poly_quad = cue.SegmentedPolynomial(
        [stp_quad.operands[0]],  # single input
        [stp_quad.operands[-1]],  # single output
        [((0, 0, 1), stp_quad)],  # use input 0 for both positions
    )

    # Test data
    x_quad = np.random.rand(16)  # 2 segments × 8 = 16

    # Original evaluation
    [result_quad_original] = poly_quad(x_quad)

    # ========================================
    # Test 2a: Split quadratic by size
    # ========================================

    # Split at segment boundary (size 8)
    poly_quad_split_size = poly_quad.split_operand_by_size(0, [0, 8, 16])

    # Now we have 2 inputs (the split parts)
    assert poly_quad_split_size.num_inputs == 2

    # Evaluate with split inputs
    [result_quad_split_size] = poly_quad_split_size(
        x_quad[0:8],  # first segment
        x_quad[8:16],  # second segment
    )

    # Results should match
    np.testing.assert_allclose(result_quad_original, result_quad_split_size)

    # ========================================
    # Test 2b: Split quadratic by segment
    # ========================================

    # Split into individual segments
    poly_quad_split_seg = poly_quad.split_operand_by_segment(0, [0, 1, 2])

    # Now we have 2 inputs (one per segment)
    assert poly_quad_split_seg.num_inputs == 2

    # Evaluate with split inputs
    [result_quad_split_seg] = poly_quad_split_seg(
        x_quad[0:8],  # segment 0
        x_quad[8:16],  # segment 1
    )

    # Results should match
    np.testing.assert_allclose(result_quad_original, result_quad_split_seg)

    # ========================================
    # Example 3: Demonstrating alignment of size and segment splits
    # ========================================

    # When we split at segment boundaries, size and segment methods give same result
    x_test = np.random.rand(64)
    y_test = np.random.rand(48)

    # Split at segment boundary (32 = 2 segments × 16)
    poly_by_size = poly.split_operand_by_size(0, [0, 32, 64])
    poly_by_segment = poly.split_operand_by_segment(0, [0, 2, 4])

    # Both should produce same structure
    assert poly_by_size.num_inputs == 3
    assert poly_by_segment.num_inputs == 3

    # Evaluate both
    [result_by_size] = poly_by_size(x_test[:32], x_test[32:], y_test)
    [result_by_segment] = poly_by_segment(x_test[:32], x_test[32:], y_test)

    # Results should be identical
    np.testing.assert_allclose(result_by_size, result_by_segment)



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

import numpy as np

import cuequivariance as cue


def fixed_axis_angle_rotation(
    irreps: cue.Irreps, axis: np.ndarray, angle: float
) -> cue.EquivariantPolynomial:
    """
    subsrcipts: ``input[u],output[u]``

    Rotation by a known angle around a known axis
    """
    assert irreps.irrep_class in [cue.SO3, cue.O3]

    d = cue.SegmentedTensorProduct.from_subscripts("iu_ju+ij")

    for mul, ir in irreps:
        # Note the transpose
        d.add_path(
            None,
            None,
            c=ir.rotation(axis, angle).T,
            dims={"u": mul},
        )

    d = d.flatten_coefficient_modes()
    return cue.EquivariantPolynomial(
        [cue.IrrepsAndLayout(irreps, cue.ir_mul)],
        [cue.IrrepsAndLayout(irreps, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(d),
    )


def yxy_rotation(
    irreps: cue.Irreps, lmax: Optional[int] = None
) -> cue.EquivariantPolynomial:
    """
    subsrcipts: ``gamma[],beta[],alpha[],input[u],output[u]``

    The composition of three rotations:

    - Rotation around the y-axis by angle gamma
    - followed by rotation around the x-axis by angle beta
    - followed by rotation around the y-axis by angle alpha

    The angles are encoded in the following way::

        encoding(x) = [cos(x * l), cos(x * (l - 1)), ..., cos(x), 1, sin(x), sin(2 * x), ..., sin(l * x)]

    where l is the maximum L in the input and output irreps.
    """
    # gamma, beta, input, A
    cbio = xy_rotation(irreps, lmax).polynomial.operations[0][1]
    aio = y_rotation(irreps, lmax).polynomial.operations[0][1]  # alpha, A, output
    cbiao = cue.segmented_polynomials.dot(
        cbio, aio, (3, 1)
    )  # gamma, beta, input, alpha, output
    cbaio = cbiao.move_operand(2, 3)  # gamma, beta, alpha, input, output
    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps.new_scalars(cbaio.operands[0].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps.new_scalars(cbaio.operands[1].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps.new_scalars(cbaio.operands[2].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(irreps, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(cbaio),
    )


def xy_rotation(
    irreps: cue.Irreps, lmax: Optional[int] = None
) -> cue.EquivariantPolynomial:
    """
    subsrcipts: ``gamma[],beta[],input[u],output[u]``

    Rotation around the y-axis followed by rotation around the x-axis
    """
    cio = y_rotation(irreps, lmax).polynomial.operations[0][1]  # gamma, input, A
    bio = x_rotation(irreps, lmax).polynomial.operations[0][1]  # beta, A, output
    cibo = cue.segmented_polynomials.dot(cio, bio, (2, 1))  # gamma, input, beta, output
    cbio = cibo.move_operand(1, 2)  # gamma, beta, input, output
    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps.new_scalars(cbio.operands[0].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps.new_scalars(cbio.operands[1].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(irreps, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(cbio),
    )


def yx_rotation(
    irreps: cue.Irreps, lmax: Optional[int] = None
) -> cue.EquivariantPolynomial:
    """
    subsrcipts: ``phi[],theta[],input[u],output[u]``

    Rotation around the x-axis followed by rotation around the y-axis
    """
    cio = x_rotation(irreps, lmax).d
    bio = y_rotation(irreps, lmax).d
    cibo = cue.segmented_polynomials.dot(cio, bio, (2, 1))
    cbio = cibo.move_operand(1, 2)
    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps.new_scalars(cbio.operands[0].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps.new_scalars(cbio.operands[1].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(irreps, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(cbio),
    )


def y_rotation(
    irreps: cue.Irreps, lmax: Optional[int] = None
) -> cue.EquivariantPolynomial:
    """
    subsrcipts: ``phi[],input[u],output[u]``

    Rotation around the y-axis by angle gamma

    The angle is encoded in the following way::

        encoding(x) = [cos(x * l), cos(x * (l - 1)), ..., cos(x), 1, sin(x), sin(2 * x), ..., sin(l * x)]

    where l is the maximum L in the input and output irreps.
    """
    assert irreps.irrep_class in [cue.SO3, cue.O3]

    if lmax is None:
        lmax = max(ir.l for _, ir in irreps)

    d = cue.SegmentedTensorProduct.from_subscripts("i_ju_ku+ijk")
    phc = d.add_segment(
        0, (lmax,)
    )  # cos(th * lmax), cos(th * (lmax - 1)), ..., cos(th)
    ph1 = d.add_segment(0, (1,))  # 1
    phs = d.add_segment(0, (lmax,))  # sin(th), sin(2 * th), ..., sin(lmax * th)

    for mul, ir in irreps:
        d.add_segment(1, (ir.l, mul))
        d.add_segment(1, (1, mul))
        d.add_segment(1, (ir.l, mul))
        sup = d.add_segment(2, (ir.l, mul))
        smi = d.add_segment(2, (1, mul))
        slo = d.add_segment(2, (ir.l, mul))

        ell = ir.l

        d.add_path(ph1, smi, smi, c=np.ones((1, 1, 1)))

        c = np.zeros((lmax, ell, ell))
        for i in range(ell):
            c[lmax - ell + i, i, i] = 1
        d.add_path(phc, sup, sup, c=c)

        c = np.zeros((lmax, ell, ell))
        for i in range(ell):
            c[ell - 1 - i, ell - 1 - i, i] = 1
        d.add_path(phs, slo, sup, c=c)

        c = np.zeros((lmax, ell, ell))
        for i in range(ell):
            c[ell - 1 - i, i, ell - 1 - i] = -1
        d.add_path(phs, sup, slo, c=c)

        c = np.zeros((lmax, ell, ell))
        for i in range(ell):
            c[lmax - ell + i, ell - 1 - i, ell - 1 - i] = 1
        d.add_path(phc, slo, slo, c=c)

    d = d.flatten_coefficient_modes()
    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps.new_scalars(d.operands[0].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(irreps, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(d),
    )


def x_rotation(
    irreps: cue.Irreps, lmax: Optional[int] = None
) -> cue.EquivariantPolynomial:
    """
    subsrcipts: ``phi[],input[u],output[u]``

    Rotation around the x-axis by angle beta

    The angle is encoded in the following way::

        encoding(x) = [cos(x * l), cos(x * (l - 1)), ..., cos(x), 1, sin(x), sin(2 * x), ..., sin(l * x)]

    where l is the maximum L in the input and output irreps.
    """
    assert irreps.irrep_class in [cue.SO3, cue.O3]

    dy = y_rotation(irreps, lmax).polynomial.operations[0][1]
    dz90 = fixed_axis_angle_rotation(
        irreps, np.array([0.0, 0.0, 1.0]), np.pi / 2.0
    ).polynomial.operations[0][1]
    d = cue.segmented_polynomials.dot(
        cue.segmented_polynomials.dot(dy, dz90, (1, 1)), dz90, (1, 1)
    )

    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps.new_scalars(d.operands[0].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(irreps, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(d),
    )


def inversion(irreps: cue.Irreps) -> cue.EquivariantPolynomial:
    """
    subsrcipts: ``input[u],output[u]``
    """
    d = cue.SegmentedTensorProduct.from_subscripts("iu_ju+ji")
    for mul, ir in irreps:
        assert len(ir.H) == 1
        H = ir.H[0]
        assert np.allclose(H @ H, np.eye(ir.dim), atol=1e-6)
        d.add_path(None, None, c=H, dims={"u": mul})
    d = d.flatten_coefficient_modes()
    return cue.EquivariantPolynomial(
        [cue.IrrepsAndLayout(irreps, cue.ir_mul)],
        [cue.IrrepsAndLayout(irreps, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(d),
    )



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import cache

import cuequivariance as cue


def symmetric_contraction(
    irreps_in: cue.Irreps,
    irreps_out: cue.Irreps,
    degrees: tuple[int, ...],
) -> cue.EquivariantPolynomial:
    """Construct the descriptor for a symmetric contraction.

    The symmetric contraction is a weighted sum of the input contracted with itself degree times.

    Subscripts: ``weights[u],input[u],output[u]``

    Args:
        irreps_in (Irreps): The input irreps, the multiplicity are treated in parallel.
        irreps_out (Irreps): The output irreps.
        degrees (tuple[int, ...]): List of degrees for the symmetric contractions.

    Returns:
        EquivariantPolynomial: The descriptor of the symmetric contraction.
            The operands are the weights, the input degree times and the output.

    Example:
        >>> cue.descriptors.symmetric_contraction(
        ...    16 * cue.Irreps("SO3", "0 + 1 + 2"),
        ...    16 * cue.Irreps("SO3", "0 + 1"),
        ...    (1, 2, 3)
        ... )
        ╭ a=32x0+80x0+176x0 b=16x0+16x1+16x2 -> C=16x0+16x1
        │  []·a[u]·b[u]➜C[u] ─────────── num_paths=4 u=16
        │  []·a[u]·b[u]·b[u]➜C[u] ────── num_paths=37 u=16
        ╰─ []·a[u]·b[u]·b[u]·b[u]➜C[u] ─ num_paths=437 u=16

        Where ``32x0+80x0+176x0`` are the weights needed for each degree (32 for degree 1, 80 for degree 2, 176 for degree 3).
    """
    return symmetric_contraction_cached(irreps_in, irreps_out, tuple(degrees))


@cache
def symmetric_contraction_cached(
    irreps_in: cue.Irreps,
    irreps_out: cue.Irreps,
    degrees: tuple[int, ...],
) -> cue.EquivariantPolynomial:
    degrees = list(degrees)
    if len(degrees) != 1:
        return cue.EquivariantPolynomial.stack(
            [
                symmetric_contraction(irreps_in, irreps_out, (degree,))
                for degree in degrees
            ],
            [True, False, False],
        )
    [degree] = degrees
    del degrees

    mul = irreps_in.muls[0]
    assert all(mul == m for m in irreps_in.muls)
    assert all(mul == m for m in irreps_out.muls)
    irreps_in = irreps_in.set_mul(1)
    irreps_out = irreps_out.set_mul(1)

    input_operands = range(1, degree + 1)
    output_operand = degree + 1

    input_operand = cue.SegmentedOperand(ndim=1, segments=[(mul,)] * irreps_in.dim)

    if degree == 0:
        d = cue.SegmentedTensorProduct.from_subscripts("i_i")
        for _, ir in irreps_out:
            if not ir.is_scalar():
                d.add_segment(output_operand, {"i": ir.dim})
            else:
                d.add_path(None, None, c=1, dims={"i": ir.dim})
        d = d.flatten_modes("i")

    else:
        abc = "abcdefgh"[:degree]
        d = cue.SegmentedTensorProduct.from_subscripts(
            f"w_{'_'.join(f'{a}' for a in abc)}_i+{abc}iw"
        )

        for i in input_operands:
            d.add_segment(i, (irreps_in.dim,))

        U = cue.reduced_symmetric_tensor_product_basis(
            irreps_in, degree, keep_ir=irreps_out, layout=cue.ir_mul
        )
        for _, ir in irreps_out:
            u = U.filter(keep=ir)
            if len(u.segments) == 0:
                d.add_segment(output_operand, {"i": ir.dim})
            else:
                [u] = u.segments  # (a, b, c, ..., i, w)
                d.add_path(None, *(0,) * degree, None, c=u)

        d = d.normalize_paths_for_operand(output_operand)
        d = d.flatten_coefficient_modes()

    d = d.append_modes_to_all_operands("u", {"u": mul})
    for i in input_operands:
        assert d.operands[i] == input_operand

    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps_in.new_scalars(d.operands[0].size), cue.ir_mul),
            cue.IrrepsAndLayout(mul * irreps_in, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(mul * irreps_out, cue.ir_mul)],
        cue.SegmentedPolynomial(
            [d.operands[0], input_operand],
            [d.operands[-1]],
            [(cue.Operation([0] + [1] * degree + [2]), d)],
        ),
    )



# SPDX-FileCopyrightText: Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
from typing import Optional, Sequence

import cuequivariance as cue
from cuequivariance.group_theory.irreps_array.irrep_utils import into_list_of_irrep


def fully_connected_tensor_product(
    irreps1: cue.Irreps, irreps2: cue.Irreps, irreps3: cue.Irreps
) -> cue.EquivariantPolynomial:
    """
    subscripts: ``weights[uvw],lhs[iu],rhs[jv],output[kw]``

    Construct a fully connected tensor product descriptor.

    The descriptor is constructed by iterating over all possible combinations of irreducible representations
    of the inputs and output irreps and constructing a path for each combination.

    .. currentmodule:: cuequivariance

    Args:
        irreps1 (Irreps): Irreps of the first operand.
        irreps2 (Irreps): Irreps of the second operand.
        irreps3 (Irreps): Irreps of the output.

    Returns:
        :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Descriptor of the fully connected tensor product.

    Examples:
        >>> cue.descriptors.fully_connected_tensor_product(
        ...    16 * cue.Irreps("SO3", "0 + 1 + 2"),
        ...    16 * cue.Irreps("SO3", "0 + 1 + 2"),
        ...    16 * cue.Irreps("SO3", "0 + 1 + 2"),
        ... )
        ╭ a=61440x0 b=16x0+16x1+16x2 c=16x0+16x1+16x2 -> D=16x0+16x1+16x2
        ╰─ [ijk]·a[uvw]·b[iu]·c[jv]➜D[kw] ─ num_paths=15 i={1, 3, 5} j={1, 3, 5} k={1, 3, 5} u=16 v=16 w=16

        Where ``61440x0`` are the 61440 weights needed to mix all the inputs with all the outputs.
    """
    G = irreps1.irrep_class

    d = cue.SegmentedTensorProduct.from_subscripts("uvw,iu,jv,kw+ijk")

    for mul, ir in irreps1:
        d.add_segment(1, (ir.dim, mul))
    for mul, ir in irreps2:
        d.add_segment(2, (ir.dim, mul))
    for mul, ir in irreps3:
        d.add_segment(3, (ir.dim, mul))

    for (i1, (mul1, ir1)), (i2, (mul2, ir2)), (i3, (mul3, ir3)) in itertools.product(
        enumerate(irreps1), enumerate(irreps2), enumerate(irreps3)
    ):
        if ir3 not in ir1 * ir2:
            continue

        # for loop over the different solutions of the Clebsch-Gordan decomposition
        for cg in G.clebsch_gordan(ir1, ir2, ir3):
            d.add_path((mul1, mul2, mul3), i1, i2, i3, c=cg)

    d = d.normalize_paths_for_operand(-1)
    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps1.new_scalars(d.operands[0].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps1, cue.ir_mul),
            cue.IrrepsAndLayout(irreps2, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(irreps3, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(d),
    )


def full_tensor_product(
    irreps1: cue.Irreps,
    irreps2: cue.Irreps,
    irreps3_filter: Optional[Sequence[cue.Irrep]] = None,
) -> cue.EquivariantPolynomial:
    """
    subscripts: ``lhs[iu],rhs[jv],output[kuv]``

    Construct a weightless channelwise tensor product descriptor.

    .. currentmodule:: cuequivariance

    Args:
        irreps1 (Irreps): Irreps of the first operand.
        irreps2 (Irreps): Irreps of the second operand.
        irreps3_filter (sequence of Irrep, optional): Irreps of the output to consider.

    Returns:
        :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Descriptor of the full tensor product.
    """
    G = irreps1.irrep_class

    if irreps3_filter is not None:
        irreps3_filter = into_list_of_irrep(G, irreps3_filter)

    d = cue.SegmentedTensorProduct.from_subscripts("iu,jv,kuv+ijk")

    for mul, ir in irreps1:
        d.add_segment(0, (ir.dim, mul))
    for mul, ir in irreps2:
        d.add_segment(1, (ir.dim, mul))

    irreps3 = []

    for (i1, (mul1, ir1)), (i2, (mul2, ir2)) in itertools.product(
        enumerate(irreps1), enumerate(irreps2)
    ):
        for ir3 in ir1 * ir2:
            if irreps3_filter is not None and ir3 not in irreps3_filter:
                continue

            # for loop over the different solutions of the Clebsch-Gordan decomposition
            for cg in cue.clebsch_gordan(ir1, ir2, ir3):
                d.add_path(i1, i2, None, c=cg)

                irreps3.append((mul1 * mul2, ir3))

    irreps3 = cue.Irreps(G, irreps3)
    irreps3, perm, inv = irreps3.sort()
    d = d.permute_segments(2, inv)

    d = d.normalize_paths_for_operand(-1)
    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps1, cue.ir_mul),
            cue.IrrepsAndLayout(irreps2, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(irreps3, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(d),
    )


def channelwise_tensor_product(
    irreps1: cue.Irreps,
    irreps2: cue.Irreps,
    irreps3_filter=None,
    simplify_irreps3: bool = False,
) -> cue.EquivariantPolynomial:
    """
    subscripts: ``weights[uv],lhs[iu],rhs[jv],output[kuv]``

    Construct a channelwise tensor product descriptor.

    This operation is computationally sparser than the fully connected tensor product.

    .. currentmodule:: cuequivariance

    Args:
        irreps1 (Irreps): Irreps of the first operand.
        irreps2 (Irreps): Irreps of the second operand.
        irreps3_filter (sequence of Irrep, optional): Irreps of the output to consider.
        simplify_irreps3 (bool, optional): If True, the irreps of the output are simplified.

    Returns:
        :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Descriptor of the channelwise tensor product.
    """
    G = irreps1.irrep_class

    if irreps3_filter is not None:
        irreps3_filter = into_list_of_irrep(G, irreps3_filter)

    d = cue.SegmentedTensorProduct.from_subscripts("uv,iu,jv,kuv+ijk")

    for mul, ir in irreps1:
        d.add_segment(1, (ir.dim, mul))
    for mul, ir in irreps2:
        d.add_segment(2, (ir.dim, mul))

    irreps3 = []

    for (i1, (mul1, ir1)), (i2, (mul2, ir2)) in itertools.product(
        enumerate(irreps1), enumerate(irreps2)
    ):
        for ir3 in ir1 * ir2:
            if irreps3_filter is not None and ir3 not in irreps3_filter:
                continue

            # for loop over the different solutions of the Clebsch-Gordan decomposition
            for cg in cue.clebsch_gordan(ir1, ir2, ir3):
                d.add_path(None, i1, i2, None, c=cg, dims={"u": mul1, "v": mul2})

                irreps3.append((mul1 * mul2, ir3))

    irreps3 = cue.Irreps(G, irreps3)
    irreps3, perm, inv = irreps3.sort()
    d = d.permute_segments(0, inv)
    d = d.permute_segments(3, inv)

    d = d.normalize_paths_for_operand(-1)

    if simplify_irreps3:
        # We need to flatten the coefficients to do this simplification
        d = d.flatten_coefficient_modes()

        # Sort output segments to regroup them by irreps
        segments = [(ir, k) for _mul, ir in irreps3 for k in range(ir.dim)]
        segments = [(ir, k, sid) for sid, (ir, k) in enumerate(segments)]
        segments = sorted(segments)
        d = d.permute_segments(3, [sid for _, _, sid in segments])
        irreps3 = irreps3.simplify()

    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps1.new_scalars(d.operands[0].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps1, cue.ir_mul),
            cue.IrrepsAndLayout(irreps2, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(irreps3, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(d),
    )


def _align_two_irreps(
    irreps1: cue.Irreps, irreps2: cue.Irreps, layout: cue.IrrepsLayout
) -> tuple[cue.Irreps, cue.Irreps]:
    assert irreps1.num_irreps == irreps2.num_irreps

    l1 = list(irreps1)
    l2 = list(irreps2)

    i = 0
    while i < min(len(l1), len(l2)):
        mul_1, ir_1 = l1[i]
        mul_2, ir_2 = l2[i]

        if mul_1 < mul_2:
            assert ir_2.dim == 1 or layout == cue.mul_ir
            l2[i] = (mul_1, ir_2)
            l2.insert(i + 1, (mul_2 - mul_1, ir_2))

        if mul_2 < mul_1:
            assert ir_1.dim == 1 or layout == cue.mul_ir
            l1[i] = (mul_2, ir_1)
            l1.insert(i + 1, (mul_1 - mul_2, ir_1))

        i += 1

    assert [mul for mul, _ in l1] == [mul for mul, _ in l2], (l1, l2)
    return cue.Irreps(irreps1.irrep_class, l1), cue.Irreps(irreps2.irrep_class, l2)


def elementwise_tensor_product(
    irreps1: cue.Irreps,
    irreps2: cue.Irreps,
    irreps3_filter: Optional[Sequence[cue.Irrep]] = None,
) -> cue.EquivariantPolynomial:
    """
    subscripts: ``lhs[iu],rhs[ju],output[ku]``

    Construct an elementwise tensor product descriptor.

    Args:
        irreps1 (Irreps): Irreps of the first operand.
        irreps2 (Irreps): Irreps of the second operand.
        irreps3_filter (sequence of Irrep, optional): Irreps of the output to consider.

    Returns:
        :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Descriptor of the elementwise tensor product.
    """
    G = irreps1.irrep_class

    if irreps1.num_irreps != irreps2.num_irreps:
        raise ValueError(
            f"The input irreps must have the same number of irreps, got {irreps1} and {irreps2}"
        )

    irreps1_cut, irreps2_cut = _align_two_irreps(irreps1, irreps2, cue.ir_mul)

    d = cue.SegmentedTensorProduct.from_subscripts("iu,ju,ku+ijk")

    irreps3 = []
    for (mul, ir1), (_, ir2) in zip(irreps1_cut, irreps2_cut):
        i1 = d.add_segment(0, (ir1.dim, mul))
        i2 = d.add_segment(1, (ir2.dim, mul))

        for ir3 in ir1 * ir2:
            if irreps3_filter is not None and ir3 not in irreps3_filter:
                continue

            for cg in G.clebsch_gordan(ir1, ir2, ir3):
                d.add_path(i1, i2, None, c=cg)

                irreps3.append((mul, ir3))

    irreps3 = cue.Irreps(G, irreps3)
    d = d.normalize_paths_for_operand(-1)
    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps1, cue.ir_mul),
            cue.IrrepsAndLayout(irreps2, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(irreps3, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(d),
    )


def linear(irreps_in: cue.Irreps, irreps_out: cue.Irreps) -> cue.EquivariantPolynomial:
    """
    subscripts: ``weights[uv],input[iu],output[iv]``

    Construct the descriptor of a linear equivariant transformation.

    Args:
        irreps_in (Irreps): Irreps of the input.
        irreps_out (Irreps): Irreps of the output.

    Returns:
        :class:`cue.EquivariantPolynomial <cuequivariance.EquivariantPolynomial>`: Descriptor of the linear transformation.
    """
    d = cue.SegmentedTensorProduct.from_subscripts("uv_iu_iv")
    for mul, ir in irreps_in:
        d.add_segment(1, (ir.dim, mul))
    for mul, ir in irreps_out:
        d.add_segment(2, (ir.dim, mul))

    for (i1, (mul1, ir1)), (i2, (mul2, ir2)) in itertools.product(
        enumerate(irreps_in), enumerate(irreps_out)
    ):
        if ir1 == ir2:
            d.add_path((mul1, mul2), i1, i2, c=1.0)

    d = d.normalize_paths_for_operand(-1)

    return cue.EquivariantPolynomial(
        [
            cue.IrrepsAndLayout(irreps_in.new_scalars(d.operands[0].size), cue.ir_mul),
            cue.IrrepsAndLayout(irreps_in, cue.ir_mul),
        ],
        [cue.IrrepsAndLayout(irreps_out, cue.ir_mul)],
        cue.SegmentedPolynomial.eval_last_operand(d),
    )
